# Copyright (C) 2024 qBraid
# Copyright (C) 2022 The Cirq Developers
#
# This file is part of the qBraid-SDK.
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. This specific file, adapted from Cirq, is dual-licensed under both the
# Apache License, Version 2.0, and the GPL v3. You may not use this file except in
# compliance with the applicable license. You may obtain a copy of the Apache License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file includes code adapted from Cirq (https://github.com/quantumlib/Cirq)
# with modifications by qBraid. The original copyright notice is included above.
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header

"""
Module defining qBraid Cirq QuilOutput.

"""
from __future__ import annotations

import string
from fractions import Fraction
from typing import Any, Callable, Optional, Union, cast

import cirq
import numpy as np
from cirq import ops, protocols, value

from qbraid.transpiler.conversions.qasm2.cirq_custom import RZZGate, U2Gate, U3Gate


def exponent_to_pi_string(exp: float) -> str:
    """A function for outputting a float exponent to a string in QUIL format."""

    exp_div_pi = exp / np.pi
    exponent_fraction = Fraction(exp_div_pi).limit_denominator(12)

    if abs(exponent_fraction.numerator) == 1 and exponent_fraction.denominator == 1:
        exponent = "pi" if exponent_fraction > 0 else "-pi"
    elif abs(exponent_fraction.numerator) == 1:
        exponent = (
            f"pi/{exponent_fraction.denominator}"
            if exponent_fraction > 0
            else f"-pi/{exponent_fraction.denominator}"
        )
    elif exponent_fraction.numerator == 0:
        exponent = "0"
    else:
        exponent = f"{exponent_fraction.numerator}*pi/{exponent_fraction.denominator}"

    return str(exponent)


def to_quil_complex_format(num) -> str:
    """A function for outputting a number to a complex string in QUIL format."""
    cnum = complex(str(num))
    return f"{cnum.real}+{cnum.imag}i"


class QuilFormatter(string.Formatter):
    """A unique formatter to correctly output values to QUIL."""

    def __init__(
        self, qubit_id_map: dict[cirq.Qid, str], measurement_id_map: dict[str, str]
    ) -> None:
        """Inits QuilFormatter.

        Args:
            qubit_id_map: A dictionary {qubit, quil_output_string} for
            the proper QUIL output for each qubit.
            measurement_id_map: A dictionary {measurement_key,
            quil_output_string} for the proper QUIL output for each
            measurement key.
        """
        self.qubit_id_map = {} if qubit_id_map is None else qubit_id_map
        self.measurement_id_map = {} if measurement_id_map is None else measurement_id_map

    def format_field(  # pylint: disable=arguments-renamed
        self, value: Any, spec: str  # pylint: disable=redefined-outer-name
    ) -> str:
        if isinstance(value, cirq.ops.Qid):
            value = self.qubit_id_map[value]
        if isinstance(value, str) and spec == "meas":
            value = self.measurement_id_map[value]
            spec = ""
        return super().format_field(value, spec)


# pylint: disable=abstract-method
@value.value_equality(approximate=True)
class QuilOneQubitGate(ops.Gate):
    """A QUIL gate representing any single qubit unitary with a DEFGATE and
    2x2 matrix in QUIL.
    """

    def __init__(self, matrix: np.ndarray) -> None:
        """Inits QuilOneQubitGate.

        Args:
            matrix: The 2x2 unitary matrix for this gate.
        """
        self.matrix = matrix

    def _num_qubits_(self) -> int:
        return 1

    def __repr__(self) -> str:
        return f"cirq.circuits.quil_output.QuilOneQubitGate(matrix=\n{self.matrix}\n)"

    def _value_equality_values_(self):
        return self.matrix


@value.value_equality(approximate=True)
class QuilTwoQubitGate(ops.Gate):
    """A two qubit gate represented in QUIL with a DEFGATE and it's 4x4
    unitary matrix.
    """

    def __init__(self, matrix: np.ndarray) -> None:
        """Inits QuilTwoQubitGate.

        Args:
            matrix: The 4x4 unitary matrix for this gate.
        """
        self.matrix = matrix

    def _num_qubits_(self) -> int:
        return 2

    def _value_equality_values_(self):
        return self.matrix

    def __repr__(self) -> str:
        return f"cirq.circuits.quil_output.QuilTwoQubitGate(matrix=\n{self.matrix}\n)"


def _ccnotpow_gate(op: cirq.Operation, formatter: QuilFormatter) -> Optional[str]:
    gate = cast(cirq.CCNotPowGate, op.gate)
    if gate._exponent != 1:
        return None
    return formatter.format("CCNOT {0} {1} {2}\n", op.qubits[0], op.qubits[1], op.qubits[2])


def _cczpow_gate(op: cirq.Operation, formatter: QuilFormatter) -> Optional[str]:
    gate = cast(cirq.CCZPowGate, op.gate)
    if gate._exponent != 1:
        return None
    lines = [
        formatter.format("H {0}\n", op.qubits[2]),
        formatter.format("CCNOT {0} {1} {2}\n", op.qubits[0], op.qubits[1], op.qubits[2]),
        formatter.format("H {0}\n", op.qubits[2]),
    ]
    return "".join(lines)


def _cnotpow_gate(op: cirq.Operation, formatter: QuilFormatter) -> Optional[str]:
    gate = cast(cirq.CNotPowGate, op.gate)
    if gate._exponent == 1:
        return formatter.format("CNOT {0} {1}\n", op.qubits[0], op.qubits[1])
    return None


def _cswap_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    return formatter.format("CSWAP {0} {1} {2}\n", op.qubits[0], op.qubits[1], op.qubits[2])


def _czpow_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(cirq.CZPowGate, op.gate)
    if gate._exponent == 1:
        return formatter.format("CZ {0} {1}\n", op.qubits[0], op.qubits[1])
    return formatter.format(
        "CPHASE({0}) {1} {2}\n",
        gate._exponent * np.pi,
        op.qubits[0],
        op.qubits[1],
    )


def _hpow_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(cirq.HPowGate, op.gate)
    if gate._exponent == 1:
        return formatter.format("H {0}\n", op.qubits[0])
    return formatter.format(
        "RY({0}) {3}\nRX({1}) {3}\nRY({2}) {3}\n",
        0.25 * np.pi,
        gate._exponent * np.pi,
        -0.25 * np.pi,
        op.qubits[0],
    )


def _identity_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    return "".join(formatter.format("I {0}\n", qubit) for qubit in op.qubits)


def _iswappow_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(cirq.ISwapPowGate, op.gate)
    if gate._exponent == 1:
        return formatter.format("ISWAP {0} {1}\n", op.qubits[0], op.qubits[1])
    return formatter.format("XY({0}) {1} {2}\n", gate._exponent * np.pi, op.qubits[0], op.qubits[1])


def _measurement_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(cirq.MeasurementGate, op.gate)
    invert_mask = gate.invert_mask
    if len(invert_mask) < len(op.qubits):
        invert_mask = invert_mask + (False,) * (len(op.qubits) - len(invert_mask))
    lines = []
    for i, (qubit, inv) in enumerate(zip(op.qubits, invert_mask)):
        if inv:
            lines.append(formatter.format("X {0} # Inverting for following measurement\n", qubit))
        lines.append(formatter.format("MEASURE {0} {1:meas}[{2}]\n", qubit, gate.key, i))
    return "".join(lines)


def _quilonequbit_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(QuilOneQubitGate, op.gate)
    return (
        f"DEFGATE USERGATE:\n    "
        f"{to_quil_complex_format(gate.matrix[0, 0])}, "
        f"{to_quil_complex_format(gate.matrix[0, 1])}\n    "
        f"{to_quil_complex_format(gate.matrix[1, 0])}, "
        f"{to_quil_complex_format(gate.matrix[1, 1])}\n"
        f'{formatter.format("USERGATE {0}", op.qubits[0])}\n'
    )


def _quiltwoqubit_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(QuilOneQubitGate, op.gate)
    return (
        f"DEFGATE USERGATE:\n    "
        f"{to_quil_complex_format(gate.matrix[0, 0])}, "
        f"{to_quil_complex_format(gate.matrix[0, 1])}, "
        f"{to_quil_complex_format(gate.matrix[0, 2])}, "
        f"{to_quil_complex_format(gate.matrix[0, 3])}\n    "
        f"{to_quil_complex_format(gate.matrix[1, 0])}, "
        f"{to_quil_complex_format(gate.matrix[1, 1])}, "
        f"{to_quil_complex_format(gate.matrix[1, 2])}, "
        f"{to_quil_complex_format(gate.matrix[1, 3])}\n    "
        f"{to_quil_complex_format(gate.matrix[2, 0])}, "
        f"{to_quil_complex_format(gate.matrix[2, 1])}, "
        f"{to_quil_complex_format(gate.matrix[2, 2])}, "
        f"{to_quil_complex_format(gate.matrix[2, 3])}\n    "
        f"{to_quil_complex_format(gate.matrix[3, 0])}, "
        f"{to_quil_complex_format(gate.matrix[3, 1])}, "
        f"{to_quil_complex_format(gate.matrix[3, 2])}, "
        f"{to_quil_complex_format(gate.matrix[3, 3])}\n"
        f'{formatter.format("USERGATE {0} {1}", op.qubits[0], op.qubits[1])}\n'
    )


def _swappow_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(cirq.SwapPowGate, op.gate)
    if gate._exponent % 2 == 1:
        return formatter.format("SWAP {0} {1}\n", op.qubits[0], op.qubits[1])
    return formatter.format(
        "PSWAP({0}) {1} {2}\n", gate._exponent * np.pi, op.qubits[0], op.qubits[1]
    )


def _twoqubitdiagonal_gate(op: cirq.Operation, formatter: QuilFormatter) -> Optional[str]:
    gate = cast(cirq.TwoQubitDiagonalGate, op.gate)
    diag_angles_radians = np.asarray(gate._diag_angles_radians)
    if np.count_nonzero(diag_angles_radians) != 1:
        return None
    if diag_angles_radians[0] != 0:
        return formatter.format(
            "CPHASE00({0}) {1} {2}\n",
            exponent_to_pi_string(diag_angles_radians[0]),
            op.qubits[0],
            op.qubits[1],
        )
    if diag_angles_radians[1] != 0:
        return formatter.format(
            "CPHASE01({0}) {1} {2}\n",
            exponent_to_pi_string(diag_angles_radians[1]),
            op.qubits[0],
            op.qubits[1],
        )
    if diag_angles_radians[2] != 0:
        return formatter.format(
            "CPHASE10({0}) {1} {2}\n",
            exponent_to_pi_string(diag_angles_radians[2]),
            op.qubits[0],
            op.qubits[1],
        )
    return formatter.format(
        "CPHASE({0}) {1} {2}\n",
        exponent_to_pi_string(diag_angles_radians[3]),
        op.qubits[0],
        op.qubits[1],
    )


def _wait_gate(*args, **kwargs) -> str:  # pylint: disable=unused-argument
    return "WAIT\n"


def _xpow_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(cirq.XPowGate, op.gate)
    if gate._exponent == 1 and gate._global_shift != -0.5:
        return formatter.format("X {0}\n", op.qubits[0])
    return formatter.format("RX({0}) {1}\n", gate._exponent * np.pi, op.qubits[0])


def _xxpow_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(cirq.XPowGate, op.gate)
    if gate._exponent == 1:
        return formatter.format("X {0}\nX {1}\n", op.qubits[0], op.qubits[1])
    return formatter.format(
        "RX({0}) {1}\nRX({2}) {3}\n",
        gate._exponent * np.pi,
        op.qubits[0],
        gate._exponent * np.pi,
        op.qubits[1],
    )


def _ypow_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(cirq.YPowGate, op.gate)
    if gate._exponent == 1 and gate.global_shift != -0.5:
        return formatter.format("Y {0}\n", op.qubits[0])
    return formatter.format("RY({0}) {1}\n", gate._exponent * np.pi, op.qubits[0])


def _yypow_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(cirq.YYPowGate, op.gate)
    if gate._exponent == 1:
        return formatter.format("Y {0}\nY {1}\n", op.qubits[0], op.qubits[1])

    return formatter.format(
        "RY({0}) {1}\nRY({2}) {3}\n",
        gate._exponent * np.pi,
        op.qubits[0],
        gate._exponent * np.pi,
        op.qubits[1],
    )


def _zpow_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(cirq.ZPowGate, op.gate)
    if gate._exponent == 1 and gate.global_shift != -0.5:
        return formatter.format("Z {0}\n", op.qubits[0])
    return formatter.format("RZ({0}) {1}\n", gate._exponent * np.pi, op.qubits[0])


def _zzpow_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(cirq.ZZPowGate, op.gate)
    if gate._exponent == 1:
        return formatter.format("Z {0}\nZ {1}\n", op.qubits[0], op.qubits[1])

    return formatter.format(
        "RZ({0}) {1}\nRZ({2}) {3}\n",
        gate._exponent * np.pi,
        op.qubits[0],
        gate._exponent * np.pi,
        op.qubits[1],
    )


def _rzz_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(RZZGate, op.gate)
    if gate._rads == 1:
        return formatter.format("CZ {0} {1}\n", op.qubits[0], op.qubits[1])
    if gate._rads == 0:
        return formatter.format("I {0}\nI {1}\n", op.qubits[0], op.qubits[1])
    return formatter.format(
        "RZZ({0}) {1} {2}\n", exponent_to_pi_string(gate._rads * np.pi), op.qubits[0], op.qubits[1]
    )


def _quil_u2_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(U2Gate, op.gate)
    return formatter.format(
        "U(1.5707963267948966, {0}, {1}) {2}\n",
        gate._phi,
        gate._lam,
        op.qubits[0],
    )


def _quil_u3_gate(op: cirq.Operation, formatter: QuilFormatter) -> str:
    gate = cast(U3Gate, op.gate)
    return formatter.format(
        "U({0}, {1}, {2}) {3}\n",
        gate._theta,
        gate._phi,
        gate._lam,
        op.qubits[0],
    )


SUPPORTED_GATES = {
    ops.CCNotPowGate: _ccnotpow_gate,
    ops.CCZPowGate: _cczpow_gate,
    ops.CNotPowGate: _cnotpow_gate,
    ops.CSwapGate: _cswap_gate,
    ops.CZPowGate: _czpow_gate,
    ops.HPowGate: _hpow_gate,
    ops.IdentityGate: _identity_gate,
    ops.ISwapPowGate: _iswappow_gate,
    ops.MeasurementGate: _measurement_gate,
    QuilOneQubitGate: _quilonequbit_gate,
    QuilTwoQubitGate: _quiltwoqubit_gate,
    ops.SwapPowGate: _swappow_gate,
    ops.TwoQubitDiagonalGate: _twoqubitdiagonal_gate,
    ops.WaitGate: _wait_gate,
    ops.XPowGate: _xpow_gate,
    ops.XXPowGate: _xxpow_gate,
    ops.YPowGate: _ypow_gate,
    ops.YYPowGate: _yypow_gate,
    ops.ZPowGate: _zpow_gate,
    ops.ZZPowGate: _zzpow_gate,
    RZZGate: _rzz_gate,
    U2Gate: _quil_u2_gate,
    U3Gate: _quil_u3_gate,
}


class QuilOutput:
    """An object for passing operations and qubits then outputting them to
    QUIL format. The string representation returns the QUIL output for the
    circuit.
    """

    def __init__(self, operations: "cirq.OP_TREE", qubits: tuple["cirq.Qid", ...]) -> None:
        """Inits QuilOutput.

        Args:
            operations: A list or tuple of `cirq.OP_TREE` arguments.
            qubits: The qubits used in the operations.
        """
        self.qubits = qubits
        self.operations = tuple(cirq.ops.flatten_to_ops(operations))
        self.measurements = tuple(
            op for op in self.operations if isinstance(op.gate, ops.MeasurementGate)
        )
        self.qubit_id_map = self._generate_qubit_ids()
        self.measurement_id_map = self._generate_measurement_ids()
        self.formatter = QuilFormatter(
            qubit_id_map=self.qubit_id_map, measurement_id_map=self.measurement_id_map
        )

    def _generate_qubit_ids(self) -> dict["cirq.Qid", str]:
        return {qubit: str(i) for i, qubit in enumerate(self.qubits)}

    def _generate_measurement_ids(self) -> dict[str, str]:
        index = 0
        measurement_id_map: dict[str, str] = {}
        for op in self.operations:
            if isinstance(op.gate, ops.MeasurementGate):
                key = protocols.measurement_key_name(op)
                if key in measurement_id_map:
                    continue
                measurement_id_map[key] = f"m{index}"
                index += 1
        return measurement_id_map

    def save_to_file(self, path: Union[str, bytes, int]) -> None:
        """Write QUIL output to a file specified by path."""
        with open(path, "w", encoding="utf-8") as f:
            f.write(str(self))

    def __str__(self) -> str:
        output = []
        self._write_quil(lambda s: output.append(s))  # pylint: disable=unnecessary-lambda
        return self.rename_defgates("".join(output))

    def _op_to_maybe_quil(self, op: cirq.Operation) -> Optional[str]:
        for gate_type, supported_gate in SUPPORTED_GATES.items():
            if isinstance(op.gate, gate_type):
                quil: Callable[[cirq.Operation, QuilFormatter], Optional[str]] = supported_gate
                return quil(op, self.formatter)
        return None

    def _op_to_quil(self, op: cirq.Operation) -> str:
        quil_str = self._op_to_maybe_quil(op)
        if not quil_str:
            raise ValueError("Can't convert Operation to string")
        return quil_str

    def _write_quil(self, output_func: Callable[[str], None]) -> None:
        output_func("# Created using qBraid.\n\n")
        if len(self.measurements) > 0:
            measurements_declared: set[str] = set()
            for m in self.measurements:
                key = protocols.measurement_key_name(m)
                if key in measurements_declared:
                    continue
                measurements_declared.add(key)
                output_func(f"DECLARE {self.measurement_id_map[key]} BIT[{len(m.qubits)}]\n")
            output_func("\n")

        def keep(op: "cirq.Operation") -> bool:
            if isinstance(op.gate, tuple(SUPPORTED_GATES.keys())):
                if not self._op_to_maybe_quil(op):
                    return False
                return True
            return False

        def fallback(op):
            if len(op.qubits) not in [1, 2]:
                return NotImplemented

            mat = protocols.unitary(op, None)
            if mat is None:
                return NotImplemented

            # Following code is a safety measure
            # Could not find a gate that doesn't decompose into a gate
            # with a _quil_ implementation
            if len(op.qubits) == 1:  # pragma: no cover
                return QuilOneQubitGate(mat).on(*op.qubits)
            return QuilTwoQubitGate(mat).on(*op.qubits)  # pragma: no cover

        def on_stuck(bad_op):
            return ValueError(f"Cannot output operation as QUIL: {bad_op!r}")

        for main_op in self.operations:
            decomposed = protocols.decompose(
                main_op, keep=keep, fallback_decomposer=fallback, on_stuck_raise=on_stuck
            )

            for decomposed_op in decomposed:
                output_func(self._op_to_quil(decomposed_op))

    def rename_defgates(self, output: str) -> str:
        """A function for renaming the DEFGATEs within the QUIL output. This
        utilizes a second pass to find each DEFGATE and rename it based on
        a counter.
        """
        result = output
        defString = "DEFGATE"
        nameString = "USERGATE"
        defIdx = 0
        nameIdx = 0
        gateNum = 0
        i = 0
        while i < len(output):
            if result[i] == defString[defIdx]:
                defIdx += 1
            else:
                defIdx = 0
            if result[i] == nameString[nameIdx]:
                nameIdx += 1
            else:
                nameIdx = 0
            if defIdx == len(defString):
                gateNum += 1
                defIdx = 0
            if nameIdx == len(nameString):
                result = result[: i + 1] + str(gateNum) + result[i + 1 :]
                nameIdx = 0
                i += 1
            i += 1
        return result



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qasm2 to qasm3 transpilation

"""
import pyqasm
import pytest
from qiskit.circuit import QuantumCircuit
from qiskit.qasm2 import dumps as qasm2_dumps
from qiskit.qasm3 import loads as qasm3_loads

from qbraid._logging import logger
from qbraid.interface import circuits_allclose
from qbraid.interface.random import random_circuit
from qbraid.transpiler.conversions.qasm2.qasm2_to_qasm3 import qasm2_to_qasm3


def _generate_valid_qasm_strings(seed=42, gates_to_skip=None, num_circuits=100):
    """Returns a list of 100 random qasm2 strings
    which do not contain any of the gates in gates_to_skip

    Current list of invalid gates is ["u", "cu1", "cu2", "cu3", "rxx", "xx_minus_yy"]
    For the motivation, see discussion
    - https://github.com/Qiskit/qiskit-qasm3-import/issues/12
    - https://github.com/Qiskit/qiskit-qasm3-import/issues/11#issuecomment-1568505732
    """
    if gates_to_skip is None:
        gates_to_skip = []

    qasm_strings = []
    while len(qasm_strings) < num_circuits:
        try:
            circuit_random = random_circuit("qiskit", seed=seed)
            qasm_str = qasm2_dumps(circuit_random)
            circuit_from_qasm = QuantumCircuit.from_qasm_str(qasm_str)
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error("Invalid QASM generated by random_circuit: %s", e)
            continue

        for gate in gates_to_skip:
            if len(circuit_from_qasm.get_instructions(gate)) > 0:
                break
        else:
            qasm_strings.append(qasm_str)

    return qasm_strings


@pytest.mark.parametrize(
    "qasm2_str", _generate_valid_qasm_strings(gates_to_skip=["r", "rccx", "xx_minus_yy"])
)
def test_random_conversion_to_qasm3(qasm2_str):
    """test random gates conversion"""
    qasm3_str = pyqasm.loads(qasm2_str).to_qasm3(as_str=True)
    circuit_orig = QuantumCircuit.from_qasm_str(qasm2_str)
    circuit_test = qasm3_loads(qasm3_str)

    # ensure that the conversion is correct
    assert circuits_allclose(circuit_orig, circuit_test)


def test_invalid_conversion():
    """Test invalid conversion inputs"""
    with pytest.raises(ValueError):
        qasm2_to_qasm3(1)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for representing various noise models for quantum simulators.

"""
import re
import threading
from collections.abc import MutableMapping
from dataclasses import dataclass, field
from typing import Iterator, Optional, Union


@dataclass(frozen=True)
class NoiseModel:
    """Class representing a single noise model."""

    # pylint: disable=line-too-long
    _defaults = {
        "ideal": "The simulation is performed without any noise, representing an ideal quantum computer.",
        "depolarizing": "Applies random errors to qubits, effectively turning a pure quantum state into a mixed state.",
        "amplitude_damping": "Simulates energy loss in a quantum system, causing qubits to decay from the excited state to the ground state.",
        "phase_damping": "Represents dephasing, where the relative phase between quantum states is randomized without energy loss.",
        "bit_flip": "Randomly flips the state of qubits (i.e., from 0 to 1 or from 1 to 0) with a certain probability.",
        "phase_flip": "Randomly flips the phase of a qubit state (i.e., it applies a Z gate) with a certain probability.",
    }
    # pylint: enable=line-too-long

    name: str = field(repr=False)
    description: Optional[str] = None
    _value: str = field(init=False)

    def __post_init__(self):
        normalized_value = self._normalize(self.name)
        object.__setattr__(self, "_value", normalized_value)
        description = self.description or self._defaults.get(normalized_value)
        self._validate(normalized_value, description)
        object.__setattr__(self, "description", description)

    @classmethod
    def _validate(cls, canonical_value: str, description: Optional[str] = None):
        if not re.match(r"^[\w\s\-_]+$", canonical_value):
            raise ValueError(
                f"Invalid noise model name: '{canonical_value}'. "
                "Allowed characters are letters, digits, spaces, hyphens, and underscores."
            )
        if description and len(description) > 120:
            raise ValueError("Description must be 120 characters or fewer.")

    @staticmethod
    def _normalize(value: str) -> str:
        return re.sub(r"[\s_]+", "_", value.strip().lower())

    def __str__(self):
        return self.value

    @property
    def value(self) -> str:
        """Normalized noise model name."""
        return self._value

    def __eq__(self, other):
        if isinstance(other, NoiseModel):
            return self.value == other.value and self.description == other.description
        if isinstance(other, str):
            return self.value == self._normalize(other)
        return NotImplemented

    def __hash__(self):
        return hash((self.value, self.description))

    def __repr__(self):
        return f"NoiseModel('{self.name}')"


class NoiseModelSet(MutableMapping):
    """Class managing a registry of noise models."""

    def __init__(self):
        self._models: dict[str, NoiseModel] = {}
        self._lock = threading.Lock()

    def add(self, name: str, description: Optional[str] = None, overwrite: bool = False):
        """Add a new noise model to the registry."""
        canonical_name = NoiseModel._normalize(name)
        new_model = NoiseModel(name, description)
        with self._lock:
            existing_model = self._models.get(canonical_name)
            if existing_model and not overwrite:
                if new_model != existing_model:
                    raise ValueError(
                        f"Noise model '{canonical_name}' already exists with a "
                        "different definition. Use overwrite=True to overwrite."
                    )
            self._models[canonical_name] = new_model

    def get(self, key: str, default: Optional[NoiseModel] = None) -> NoiseModel:
        """Retrieve a noise model by name."""
        canonical_name = NoiseModel._normalize(key)
        return self._models.get(canonical_name, default)

    def remove(self, name: str):
        """Remove a noise model from the registry."""
        canonical_name = NoiseModel._normalize(name)
        with self._lock:
            if canonical_name not in self._models:
                raise KeyError(f"Noise model '{name}' not found.")
            del self._models[canonical_name]

    def discard(self, name: str):
        """Remove a noise model if it exists; do nothing otherwise."""
        canonical_name = NoiseModel._normalize(name)
        with self._lock:
            self._models.pop(canonical_name, None)

    def clear(self):
        """Clear all noise models from the registry."""
        with self._lock:
            self._models.clear()

    def update(self, other: "NoiseModelSet"):  # pylint: disable=arguments-differ
        """Update the registry with another NoiseModelSet instance."""
        if isinstance(other, dict):
            other = NoiseModelSet.from_dict(other)
        elif not isinstance(other, NoiseModelSet):
            raise TypeError("Can only update from a dict or another NoiseModelSet instance.")
        with self._lock:
            with other._lock:
                self._models.update(other._models)

    def __getitem__(self, key: str) -> NoiseModel:
        noise_model = self.get(key)
        if noise_model is None:
            raise KeyError(f"Noise model {key} not found.")
        return noise_model

    def __setitem__(self, key: str, value: NoiseModel):
        if not isinstance(value, NoiseModel):
            raise ValueError("Value must be a NoiseModel instance.")
        canonical_name = NoiseModel._normalize(key)
        if canonical_name != value.value:
            raise ValueError("Key does not match the NoiseModel's normalized value.")
        with self._lock:
            self._models[canonical_name] = value

    def __delitem__(self, key: str):
        self.remove(key)

    def __iter__(self) -> Iterator[str]:
        return iter(self._models)

    def __len__(self) -> int:
        return len(self._models)

    def __contains__(self, key: str) -> bool:
        if not isinstance(key, str):
            return False
        canonical_name = NoiseModel._normalize(key)
        return canonical_name in self._models

    def __repr__(self):
        return f"NoiseModelSet({list(self._models.keys())})"

    def values(self) -> Iterator[NoiseModel]:
        """Return an iterator over the noise models."""
        return iter(self._models.values())

    def items(self):
        """Return an iterator over the (name, noise model) pairs."""
        return self._models.items()

    @classmethod
    def from_dict(cls, data: dict[str, str]) -> "NoiseModelSet":
        """Create a NoiseModelSet instance from a dictionary."""
        models = cls()
        for name, description in data.items():
            models.add(name, description)
        return models

    @classmethod
    def from_iterable(cls, data: Union[list[str], set[str], tuple[str, ...]]) -> "NoiseModelSet":
        """Create a NoiseModelSet instance from a list, set, or tuple."""
        models = cls()
        for name in data:
            models.add(name)
        return models



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining CUDA-Q conversion extras.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

import cudaq
from qbraid_core._import import LazyLoader

from qbraid.transpiler.annotations import requires_extras

if TYPE_CHECKING:
    from cudaq import PyKernel
    from pyqir import Module

pyqir = LazyLoader("pyqir", globals(), "pyqir")


@requires_extras("pyqir")
def cudaq_to_pyqir(kernel: PyKernel) -> Module:
    """Converts a CUDA-Q kernel to PyQIR."""
    llvm_ir = cudaq.translate(kernel, format="qir")
    context = pyqir.Context()
    module = pyqir.Module.from_ir(context, llvm_ir, "main")
    return module



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Configures display options based on environment context.

"""
import sys

from ._logging import logger


def running_in_jupyter() -> bool:
    """
    Determine if running within Jupyter.

    Returns:
        bool: True if running in Jupyter, else False.
    """
    try:
        if "IPython" in sys.modules:
            get_ipython = sys.modules["IPython"].__dict__["get_ipython"]
            ip = get_ipython()
            if ip is not None:
                return getattr(ip, "kernel", None) is not None
    except Exception as err:  # pylint: disable=broad-exception-caught
        logger.error("Error checking if running in Jupyter: %s", err)

    return False



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining OpenQASM 3 to CUDA-Q conversion function.

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import pyqasm
from openqasm3 import ast
from qbraid_core._import import LazyLoader

from qbraid.transpiler.annotations import weight
from qbraid.transpiler.exceptions import ProgramConversionError

cudaq = LazyLoader("cudaq", globals(), "cudaq")

if TYPE_CHECKING:
    from cudaq import PyKernel, QuakeValue

    from qbraid.programs.typer import QasmStringType


def make_gate_kernel(name: str, targs: tuple[type]) -> PyKernel:
    """Returns CUDA-Q kernel for pure standard gates (no modifiers - ctrl or adj)."""

    if name in [
        "x",
        "y",
        "z",
        "rx",
        "ry",
        "rz",
        "h",
        "s",
        "t",
        "sdg",
        "tdg",
        "u3",
        "i",
        "id",
        "iden",
    ]:
        size = 1
    elif name in ["swap"]:
        size = 2
    else:
        raise ProgramConversionError(f"Unsupported gate: {name}")

    kernel, *qparams = cudaq.make_kernel(*[cudaq.qubit for _ in range(size)], *targs)
    qrefs, qargs = qparams[:size], qparams[size:]

    if name in ["i", "id", "iden"]:
        return kernel

    op = getattr(kernel, name)

    if len(targs) > 0:
        op(*qargs, *qrefs)
    else:
        op(*qrefs)

    return kernel


@weight(0.95)  # pylint: disable-next=too-many-statements
def openqasm3_to_cudaq(program: QasmStringType | ast.Program) -> PyKernel:
    """Returns a CUDA-Q kernel representing the input OpenQASM program.

    Args:
        qasm (str or ast.Program): OpenQASM program to convert to CUDA-Q kernel.

    Returns:
        kernel: CUDA-Q kernel equivalent to input OpenQASM string.
    """
    try:
        module = pyqasm.loads(program)
        module.validate()
    except Exception as e:
        raise ProgramConversionError("QASM program is not well-formed.") from e

    module.unroll()
    program = module.unrolled_ast

    kernel: PyKernel = cudaq.make_kernel()
    ctx: dict[str, Optional[QuakeValue]] = {}
    gate_kernels: dict[str, PyKernel] = {}

    def get_gate(name: str, targs: tuple[type]) -> PyKernel:
        if name in gate_kernels:
            return gate_kernels[name]

        gate_kernels[name] = make_gate_kernel(name, targs)
        return gate_kernels[name]

    def qubit_lookup(qubit: ast.IndexedIdentifier | ast.Identifier) -> QuakeValue:
        assert isinstance(
            qubit, ast.IndexedIdentifier
        ), f"all identifiers should've been indexed: {qubit}"

        assert len(qubit.indices) == 1, f"multi-dim arrays are not supported: {qubit.indices}"

        inds = qubit.indices[0]

        assert len(inds) == 1, f"indices should've been a single integer: {inds}"
        assert isinstance(ind := inds[0], ast.IntegerLiteral)

        q = ctx[qubit.name.name][ind.value]
        return q

    # pylint: disable-next=too-many-nested-blocks
    for statement in program.statements:
        if isinstance(statement, ast.Include):
            if statement.filename not in {"stdgates.inc", "qelib1.inc"}:
                raise ProgramConversionError(f"Custom includes are unsupported: {statement}")
        elif isinstance(statement, ast.QubitDeclaration):
            ctx[statement.qubit.name] = kernel.qalloc(statement.size.value)
        elif isinstance(statement, ast.ClassicalDeclaration):
            if statement.init_expression and isinstance(
                statement.init_expression, ast.QuantumMeasurement
            ):
                ctx[statement.identifier.name] = kernel.mz(
                    qubit_lookup(statement.init_expression.qubit)
                )
            else:
                ctx[statement.identifier.name] = None
        elif isinstance(statement, ast.QuantumMeasurementStatement):
            val = kernel.mz(qubit_lookup(statement.measure.qubit))
            if statement.target is not None:
                assert isinstance(
                    statement.target, ast.IndexedIdentifier
                ), f"identifiers should've been unrolled to indexed identifiers: {statement.target}"
                ctx[statement.target.name.name] = val
        elif isinstance(statement, ast.QuantumGate):
            name, qubits = statement.name.name, statement.qubits

            args = []
            for arg in statement.arguments:
                assert arg.value is not None, f"gate arguments should've been literals: {arg}"
                args.append(arg.value)
            targs = [type(a) for a in args]

            qubit_refs = [qubit_lookup(q) for q in qubits]

            # pyqasm unrolls multiple modifiers.
            # ctrl isn't supported so multi-ctrl is not an issue at the moment.
            assert len(statement.modifiers) <= 1

            if len(statement.modifiers) == 1:
                mod = statement.modifiers[0]
                assert (
                    mod.modifier == ast.GateModifierName.ctrl
                ), f"non-ctrl modifiers should've be unrolled: {mod}"

                gate = get_gate(name, targs)
                kernel.control(gate, qubit_refs[0], *qubit_refs[1:])
            else:
                if (namel := name.lower())[0] == "c" and namel[1:] in [
                    "x",
                    "y",
                    "z",
                    "rx",
                    "ry",
                    "rz",
                ]:
                    # pyqasm doesn't unroll C{X,Y,Z} -> ctrl @ x. the below also handles this.
                    gate = get_gate(namel[1:], targs)
                    kernel.control(gate, qubit_refs[0], *qubit_refs[1:], *args)
                else:
                    gate = get_gate(name, targs)
                    kernel.apply_call(gate, *qubit_refs, *args)

        else:
            raise ProgramConversionError(f"Unsupported statement: {statement}")

    return kernel



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit test for quantum program registry

"""
import random
import string
import sys
import unittest.mock

import pytest

from qbraid.exceptions import QbraidError
from qbraid.interface import random_circuit
from qbraid.programs import (
    QPROGRAM_ALIASES,
    QPROGRAM_REGISTRY,
    QPROGRAM_TYPES,
    ProgramSpec,
    derive_program_type_alias,
    get_program_type_alias,
    load_program,
    register_program_type,
    unregister_program_type,
)
from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.registry import get_native_experiment_type, is_registered_alias_native
from qbraid.programs.typer import IonQDict


def generate_unique_key(registry: dict, length: int = 10) -> str:
    """
    Generate a random string that is not already a key in the provided registry.

    """
    attempts = 0
    while attempts < 10:
        new_key = "".join(random.choices(string.ascii_letters + string.digits, k=length))
        if new_key not in registry:
            return new_key.lower()
        attempts += 1

    raise RuntimeError(f"Failed to generate a unique key after {attempts} attempts.")


def test_successful_registration():
    """Test successful registration of a program type"""
    try:
        register_program_type(dict, "dict")
        assert "dict" in QPROGRAM_REGISTRY
        assert QPROGRAM_REGISTRY["dict"] is dict
    finally:
        unregister_program_type("dict")


def test_registration_without_alias():
    """Test registration of a program type without an alias"""
    unregister_program_type("unittest", raise_error=False)
    try:
        register_program_type(unittest.mock.Mock)
        assert "unittest" in QPROGRAM_REGISTRY
        assert QPROGRAM_REGISTRY["unittest"] == unittest.mock.Mock
    finally:
        unregister_program_type("unittest", raise_error=False)


def test_error_on_duplicate_alias():
    """Test error when trying to register a duplicate alias"""
    register_program_type(str, "qasm2")
    with pytest.raises(ValueError) as excinfo:
        register_program_type(int, "qasm2")
    assert "Alias 'qasm2' is already registered with a different type." in str(excinfo.value)


def test_error_on_duplicate_type_different_alias():
    """Test error when trying to register a type with a different alias"""
    try:
        register_program_type(str, "custom_str")

        with pytest.raises(ValueError) as excinfo:
            register_program_type(str, "another_str")

        expected_msg = (
            "Cannot register more than one additional 'str' type beyond "
            "'qasm2', 'qasm3', and 'qasm2_kirin'."
        )
        assert expected_msg in str(excinfo.value)
    finally:
        unregister_program_type("custom_str", raise_error=False)
        unregister_program_type("another_str", raise_error=False)


def test_error_on_duplicate_type():
    """Test error when trying to register a type that is already registered"""
    register_program_type(int, "test_type_0")
    with pytest.raises(ValueError):
        register_program_type(int, "test_type_1")

    unregister_program_type("test_type_0", raise_error=False)


def test_re_registration_same_alias_type():
    """Test re-registration of the same type with the same alias"""
    init_registry_size = len(QPROGRAM_REGISTRY)
    register_program_type(str, "qasm2")
    register_program_type(str, "qasm2")
    assert len(QPROGRAM_REGISTRY) == init_registry_size


def test_get_program_type_alias_alias_library_mismatch():
    """Test getting program type when the alias does not match the module."""
    try:
        register_program_type(unittest.mock.Mock, alias="mock")
        assert get_program_type_alias(unittest.mock.Mock()) == "mock"
    finally:
        unregister_program_type("mock")


def test_overwrite_existing_alias():
    """Test overwriting an existing alias with a new type"""
    try:
        register_program_type(str, "qasm2")
        register_program_type(unittest.mock.Mock, "qasm2", overwrite=True)
        assert QPROGRAM_REGISTRY["qasm2"] == unittest.mock.Mock
    finally:
        register_program_type(str, "qasm2", overwrite=True)


def test_is_alias_registered_native_true():
    """Verify that is_alias_registered_native returns True when
    a native program type is registered with a given alias."""
    register_program_type(str, "qasm2")
    assert is_registered_alias_native("qasm2")


def test_is_alias_registered_native_false():
    """Verify that is_alias_registered_native returns False when the registered
    program type for an alias does not match the native program type."""
    try:
        register_program_type(dict, "qasm2", overwrite=True)
        assert not is_registered_alias_native("qasm2")
    finally:
        register_program_type(str, "qasm2", overwrite=True)


def test_register_via_program_spec():
    """Test registering a program type using ProgramSpec"""
    unregister_program_type("unittest", raise_error=False)
    try:
        spec = ProgramSpec(unittest.mock.Mock)
        assert "unittest" in QPROGRAM_REGISTRY
        assert spec.alias == "unittest"
        assert spec.native is False
        assert spec.program_type == unittest.mock.Mock
    finally:
        unregister_program_type("unittest", raise_error=False)


def test_unregister_program_type_error_handling():
    """Test unregistering a program type with different raise_error settings."""
    alias = generate_unique_key(QPROGRAM_REGISTRY)

    try:
        unregister_program_type(alias, raise_error=False)
    except KeyError:
        pytest.fail("KeyError was raised, but raise_error was set to False")

    with pytest.raises(KeyError):
        unregister_program_type(alias, raise_error=True)


def test_unregister_program_type_unique():
    """Test unregistering a program alias where the
    program type is unique within the registry."""
    QPROGRAM_TYPES.discard(dict)
    assert dict not in QPROGRAM_TYPES
    alias = generate_unique_key(QPROGRAM_REGISTRY)
    register_program_type(dict, alias, overwrite=True)
    assert alias in QPROGRAM_REGISTRY
    assert alias in QPROGRAM_ALIASES
    assert dict in QPROGRAM_TYPES
    unregister_program_type(alias)
    assert alias not in QPROGRAM_REGISTRY
    assert alias not in QPROGRAM_ALIASES
    assert dict not in QPROGRAM_TYPES


def test_unregister_program_type_non_unique():
    """Test unregistering a program alias where the
    program type is not unique within the registry."""
    alias = generate_unique_key(QPROGRAM_REGISTRY)
    register_program_type(str, alias, overwrite=True)
    register_program_type(str, "qasm2", overwrite=True)
    assert alias in QPROGRAM_REGISTRY
    assert alias in QPROGRAM_ALIASES
    assert str in QPROGRAM_TYPES
    unregister_program_type(alias)
    assert alias not in QPROGRAM_REGISTRY
    assert alias not in QPROGRAM_ALIASES
    assert str in QPROGRAM_TYPES


def test_program_spec_str_rep():
    """Test string representation of ProgramSpec"""
    spec = ProgramSpec(str, alias="qasm2")
    assert str(spec) == "ProgramSpec(str, qasm2)"


def test_program_spec_equality():
    """Test equality of ProgramSpec instances"""
    spec1 = ProgramSpec(str, alias="qasm2")
    spec2 = {}
    assert spec1 != spec2


class FakeType:
    """Fake type for testing automatic alias registration"""

    def __module__(self):
        return None


def test_error_on_automatic_alias():
    """Test error when trying to register a program type with an automatic alias"""
    with pytest.raises(ValueError):
        derive_program_type_alias(FakeType)


def test_derive_program_type_alias_qbraid_meta():
    """Test deriving an alias from a QbraidMetaType instance"""
    assert derive_program_type_alias(IonQDict) == "ionq"


@pytest.mark.skipif(sys.version_info < (3, 10), reason="Requires Python 3.10 or higher")
def test_load_entrypoint_not_found():
    """Test error when trying to load a program type that is not found"""
    with unittest.mock.patch("importlib.metadata.entry_points") as mock_entry_points:
        mock_entry_points.return_value.select.return_value = []
        with pytest.raises(QbraidError):
            program = random_circuit("qiskit")
            load_program(program)


def test_program_type_error():
    """Test the ProgramTypeError class."""
    error = ProgramTypeError(program="test")
    message = error.generate_message()
    assert message == "Quantum program of type '<class 'str'>' is not supported."

    error = ProgramTypeError()
    message = error.generate_message()
    assert message == "Unsupported quantum program type."


def test_get_native_experiment_type_not_found():
    """Test error when trying to get the native experiment type that is not found."""
    with pytest.raises(ValueError) as excinfo:
        get_native_experiment_type("not_found")
    assert "Entry point 'not_found' not found in 'qbraid.programs'." in str(excinfo.value)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing pyQuil programs used for testing

"""
import numpy as np
from pyquil import Program
from pyquil.gates import CNOT, CPHASE, ISWAP, RX, RY, RZ, SWAP, H, X, Y, Z


def pyquil_bell() -> Program:
    """Returns pyQuil bell circuit"""
    program = Program()
    program += H(0)
    program += CNOT(0, 1)
    return program


def pyquil_shared15() -> Program:
    """Returns pyquil `Program` for qBraid `TestSharedGates`."""

    p = Program()

    p += H(0)
    p += H(1)
    p += H(2)
    p += H(3)
    p += X(0)
    p += X(1)
    p += Y(2)
    p += Z(3)
    p += RZ(np.pi / 2, 0)  # S
    p += RZ(-np.pi / 2, 1)  # Sdg
    p += RZ(np.pi / 4, 2)  # T
    p += RZ(-np.pi / 4, 3)  # Tdg
    p += RX(np.pi / 4, 0)
    p += RY(np.pi / 2, 1)
    p += RZ(3 * np.pi / 4, 2)
    p += RZ(np.pi / 8, 3)
    p += RX(np.pi / 2, 0)
    p += RX(-np.pi / 2, 1)
    p += ISWAP(2, 3)
    p += SWAP(0, 2)
    p += SWAP(1, 3)
    p += CNOT(0, 1)
    p += CPHASE(np.pi / 4, 2, 3)

    return p



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qbraid.programs.pyquil.PyQuilProgram

"""
import pytest

try:
    from pyquil import Program

    from qbraid.programs.exceptions import ProgramTypeError
    from qbraid.programs.gate_model.pyquil import PyQuilProgram

    pyquil_not_installed = False
except ImportError:
    pyquil_not_installed = True


pytestmark = pytest.mark.skipif(pyquil_not_installed, reason="pyquil not installed")


def test_invalid_program_type():
    """Test raising ProgramTypeError for invalid program type"""
    with pytest.raises(ProgramTypeError):
        PyQuilProgram("OPENQASM 2.0;qreg q[2];h q[0];cx q[0],q[1];")


def test_program_properties():
    """Test properties of PyQuilProgram"""
    program = PyQuilProgram(Program())
    assert program.num_clbits == 0
    assert program.depth == 0



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining PytketCircuit Class

"""

from typing import Optional, Union

import numpy as np
from pytket.circuit import Circuit, Command, OpType
from pytket.circuit_library import TK1_to_RzRx
from pytket.passes import DecomposeBoxes, RebaseCustom, SafetyMode
from pytket.predicates import (
    CompilationUnit,
    GateSetPredicate,
    MaxNQubitsPredicate,
    NoClassicalControlPredicate,
    NoFastFeedforwardPredicate,
    NoMidMeasurePredicate,
    NoSymbolsPredicate,
)
from pytket.unit_id import Qubit

from qbraid.programs.exceptions import ProgramTypeError, TransformError

from ._model import GateModelProgram

IONQ_GATES = {
    OpType.X,
    OpType.Y,
    OpType.Z,
    OpType.Rx,
    OpType.Ry,
    OpType.Rz,
    OpType.H,
    OpType.CX,
    OpType.S,
    OpType.Sdg,
    OpType.T,
    OpType.Tdg,
    OpType.V,
    OpType.Vdg,
    OpType.Measure,
    OpType.noop,
    OpType.SWAP,
    OpType.ZZPhase,
    OpType.XXPhase,
    OpType.YYPhase,
    OpType.ZZMax,
    OpType.Barrier,
}


class PytketCircuit(GateModelProgram):
    """Wrapper class for ``pytket.circuit.Circuit`` objects."""

    def __init__(self, program: Circuit):
        super().__init__(program)
        if not isinstance(program, Circuit):
            raise ProgramTypeError(
                message=f"Expected 'pytket.Circuit' object, got '{type(program)}'."
            )

    @property
    def qubits(self) -> list[Qubit]:
        """Return the qubits acted upon by the operations in this circuit"""
        return self.program.qubits

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits in the circuit."""
        return self.program.n_qubits

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return self.program.n_bits

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        return self.program.depth()

    @staticmethod
    def remove_measurements(original_circuit):
        """
        Return a new circuit with all non-measurement operations from the original circuit.

        Args:
            original_circuit: The pytket Circuit to process.

        Returns:
            A new pytket Circuit without measurement operations.
        """
        new_circuit = Circuit()
        for qreg in original_circuit.qubits:
            new_circuit.add_qubit(qreg)
        for creg in original_circuit.bits:
            new_circuit.add_bit(creg)

        for command in original_circuit.get_commands():
            if command.op.type != OpType.Measure:
                new_circuit.add_gate(
                    command.op.type, command.op.params, [q.index[0] for q in command.qubits]
                )

        return new_circuit

    def _unitary(self) -> np.ndarray:
        """Return the unitary of a pytket circuit."""
        try:
            return self.program.get_unitary()
        except RuntimeError:
            program_copy = self.remove_measurements(self.program)
            return program_copy.get_unitary()

    def remove_idle_qubits(self) -> None:
        """Checks whether the circuit uses contiguous qubits/indices,
        and if not, reduces dimension accordingly."""
        circuit = self.program.copy()
        circuit.remove_blank_wires()
        self._program = circuit

    def reverse_qubit_order(self) -> None:
        """Reverses the qubit ordering of a PyTKET circuit."""
        new_c = Circuit(self.num_qubits)
        for gate in self.program.get_commands():
            circuit_qubits = gate.qubits
            circuit_qubits = [(self.num_qubits - 1) - qubits.index[0] for qubits in circuit_qubits]
            gate_op = gate.op
            # devide parameter by pi, from radian to degree
            new_c.add_gate(
                gate_op.type,
                np.array(gate_op.params) / np.pi if gate_op.params else gate_op.params,
                circuit_qubits,
            )
        self._program = new_c

    @staticmethod
    def gate_to_matrix(
        gates: Optional[Union[list[Circuit], Command]], flat: bool = False
    ) -> np.ndarray:
        """Return the unitary of the Command"""
        gates = gates if isinstance(gates, list) else [gates]
        a = list(map(max, [gate.qubits for gate in gates]))
        circuit = Circuit(max(a).index[0] + 1)
        for gate in gates:
            gate_op = gate.op
            circuit.add_gate(gate_op.type, gate_op.params, gate.qubits)
        if flat:
            circuit.remove_blank_wires()
        return circuit.get_unitary()

    @staticmethod
    def rebase(
        circuit: Circuit,
        gates: set[OpType],
        max_qubits: int,
        safety_mode: SafetyMode = SafetyMode.Default,
    ) -> Circuit:
        """
        Rebase a PyTKET circuit according to a given gate set and maximum number of qubits.

        Args:
            circuit (pytket.circuit.Circuit): The input PyTKET circuit to be rebased.

        Returns:
            pytket.circuit.Circuit: The rebased PyTKET circuit.

        """
        preds = [
            NoClassicalControlPredicate(),
            NoFastFeedforwardPredicate(),
            NoMidMeasurePredicate(),
            NoSymbolsPredicate(),
            GateSetPredicate(gates),
            MaxNQubitsPredicate(max_qubits),
        ]

        ionq_rebase_pass = RebaseCustom(
            gateset=gates, cx_replacement=Circuit(), tk1_replacement=TK1_to_RzRx
        )
        cu = CompilationUnit(circuit, preds)
        DecomposeBoxes().apply(cu, safety_mode=safety_mode)
        ionq_rebase_pass.apply(cu, safety_mode=safety_mode)
        try:
            assert cu.check_all_predicates()
        except AssertionError as err:
            raise TransformError(
                "Rebased circuit failed to satisfy compilation predicates"
            ) from err
        return cu.circuit

    def transform(self, device) -> None:
        """Transform program to according to device target profile."""
        num_qubits = device.profile.get("num_qubits")
        provider_name = device.profile.get("provider_name", "").upper()

        if provider_name == "IONQ":
            self._program = self.rebase(self.program, IONQ_GATES, num_qubits)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Information about qBraid and dependencies.

"""
import datetime
import importlib.metadata
import platform
import re
from typing import Optional

from qbraid_core._compat import check_version

from ._version import __version__


def get_dependencies(
    package_name, exclude_extras: Optional[set[str]] = None
) -> tuple[set[str], dict[str, set[str]]]:
    """
    Extracts core and optional dependencies of a package using importlib.metadata.

    Args:
        package_name (str): Name of the package to analyze.
        exclude_extras (optional, set[str]): Optional set of extras to exclude from the analysis.

    Returns:
        Tuple containing two elements:
        - set[str]: Core dependencies without any extras.
        - dict[str, set[str]]: Dependencies categorized by their extras
    """
    dist = importlib.metadata.distribution(package_name)
    requires = dist.requires or []

    core_dependencies = set()
    optional_dependencies = {}
    extras_regex = re.compile(r'^(.+?); extra == "([^"]+)"$')
    package_name_pattern = re.compile(r"^([^>=<\[\]]+)")

    exclude_extras = exclude_extras or set()

    for req in requires:
        req = req.strip()
        match = extras_regex.match(req)
        if match:
            dependency, extra = match.groups()
            if extra not in exclude_extras:
                dependency = package_name_pattern.match(dependency.strip()).group(0)
                if extra in optional_dependencies:
                    extras: set[str] = optional_dependencies[extra]
                    extras.add(dependency)
                    optional_dependencies[extra] = extras
                else:
                    optional_dependencies[extra] = {dependency}
        else:
            dependency = package_name_pattern.match(req).group(0)
            core_dependencies.add(dependency)

    return core_dependencies, optional_dependencies


def about() -> None:
    """
    Displays information about the qBraid-SDK including the installed versions
    of its core and optional dependencies, as well as the Python version and
    operating platform on which it is running.
    """
    check_version("qbraid")
    exclude_extras = {"test", "lint", "docs"}
    core_packages, extras = get_dependencies("qbraid", exclude_extras=exclude_extras)
    optional_packages = {item for subset in extras.values() for item in subset}
    all_packages = core_packages | optional_packages

    core_dependencies = {}
    optional_dependencies = {}

    for pkg in sorted(all_packages):
        try:
            version = importlib.metadata.distribution(pkg).version
            if pkg in core_packages:
                core_dependencies[pkg] = version
            elif pkg in optional_packages:
                optional_dependencies[pkg] = version
        except importlib.metadata.PackageNotFoundError:
            continue

    current_year = datetime.datetime.now().year

    about_str = (
        "\nqBraid-SDK: A platform-agnostic quantum runtime framework\n"
        "=========================================================\n"
        f"(C) {current_year} qBraid Development Team (https://sdk.qbraid.com)\n\n"
        f"qbraid:\t{__version__}\n\n"
        "Core Dependencies\n"
        "-----------------\n"
        + "\n".join([f"{k}: {v}" for k, v in core_dependencies.items()])
        + "\n\n"
        "Optional Dependencies\n"
        "---------------------\n"
    )

    if optional_dependencies:
        about_str += "\n".join([f"{k}: {v}" for k, v in optional_dependencies.items()])
    else:
        about_str += "None"

    about_str += (
        f"\n\nPython: {platform.python_version()}\n"
        f"Platform: {platform.system()} ({platform.machine()})"
    )

    print(about_str)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining PyQuilProgram Class

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from pyquil import Program
from pyquil.quilbase import Declare, Measurement
from pyquil.simulation.tools import program_unitary

from qbraid.programs.exceptions import ProgramTypeError

from ._model import GateModelProgram

if TYPE_CHECKING:
    import numpy as np
    import pyquil


class PyQuilProgram(GateModelProgram):
    """Wrapper class for ``pyQuil.Program`` objects."""

    def __init__(self, program: pyquil.Program):
        super().__init__(program)
        if not isinstance(program, Program):
            raise ProgramTypeError(
                message=f"Expected 'pyquil.Program' object, got '{type(program)}'."
            )

    @property
    def qubits(self) -> list[int]:
        """Return the qubits acted upon by the operations in this circuit"""
        return list(self.program.get_qubits())

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return 0

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        return len(self.program)

    @staticmethod
    def remove_measurements(original_program):
        """Remove MEASURE and DECLARE instructions from a pyQuil program."""
        program = Program()
        for instruction in original_program:
            if not isinstance(instruction, Measurement) and not isinstance(instruction, Declare):
                program += instruction
        return program

    def _unitary(self) -> np.ndarray:
        """Return the unitary of a pyQuil program."""
        program_copy = self.remove_measurements(self.program)
        return program_unitary(program_copy, n_qubits=self.num_qubits)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for transpiling ciruits with idle qubits

"""
import braket.circuits
import cirq
import pytest
import qiskit

from qbraid.interface.circuit_equality import assert_allclose_up_to_global_phase, circuits_allclose
from qbraid.programs import load_program
from qbraid.transpiler import transpile

# pylint: disable=redefined-outer-name


@pytest.fixture
def braket_circuit() -> braket.circuits.Circuit:
    """Returns Braket bell circuit with idle qubits"""
    return braket.circuits.Circuit().h(4).cnot(4, 8)


@pytest.fixture
def cirq_circuit() -> cirq.Circuit:
    """Returns Cirq bell circuit with idle qubits"""
    q4, q8 = cirq.LineQubit(4), cirq.LineQubit(8)
    circuit = cirq.Circuit(cirq.ops.H(q4), cirq.ops.CNOT(q4, q8))
    return circuit


@pytest.fixture
def qiskit_circuit() -> qiskit.QuantumCircuit:
    """Returns Qiskit bell circuit with idle qubits"""
    circuit = qiskit.QuantumCircuit(9)
    circuit.h(4)
    circuit.cx(4, 8)
    return circuit


def test_braket_to_cirq(braket_circuit):
    """Tests Braket conversions"""
    cirq_test = transpile(braket_circuit, "cirq", require_native=True)
    assert circuits_allclose(cirq_test, braket_circuit)


def test_braket_to_qiskit(braket_circuit):
    """Tests Braket conversions"""
    qiskit_test = transpile(braket_circuit, "qiskit", require_native=True)
    qprogram_qiskit = load_program(qiskit_test)
    qprogram_braket = load_program(braket_circuit)
    qprogram_braket.populate_idle_qubits()
    qiskit_u = qprogram_qiskit.unitary()
    braket_u = qprogram_braket.unitary()
    assert_allclose_up_to_global_phase(qiskit_u, braket_u, atol=1e-7)


def test_cirq_to_braket(cirq_circuit):
    """Tests Cirq conversions"""
    braket_test = transpile(cirq_circuit, "braket", require_native=True)
    assert circuits_allclose(braket_test, cirq_circuit)


def test_cirq_to_qiskit(cirq_circuit):
    """Tests Cirq conversions"""
    qiskit_test = transpile(cirq_circuit, "qiskit", require_native=True)
    assert circuits_allclose(qiskit_test, cirq_circuit)


def test_qiskit_to_cirq(qiskit_circuit):
    """Tests Qiskit conversions"""
    cirq_test = transpile(qiskit_circuit, "cirq", require_native=True)
    qprogram_qiskit = load_program(qiskit_circuit)
    qprogram_cirq = load_program(cirq_test)
    qprogram_cirq.populate_idle_qubits()
    qiskit_u = qprogram_qiskit.unitary()
    cirq_u = qprogram_cirq.unitary()
    assert_allclose_up_to_global_phase(qiskit_u, cirq_u, atol=1e-7)


def test_qiskit_to_braket(qiskit_circuit):
    """Tests Qiskit conversions"""
    braket_test = transpile(qiskit_circuit, "braket", require_native=True)
    qprogram_qiskit = load_program(qiskit_circuit)
    qprogram_braket = load_program(braket_test)
    qprogram_braket.populate_idle_qubits()
    qiskit_u = qprogram_qiskit.unitary()
    braket_u = qprogram_braket.unitary()
    assert_allclose_up_to_global_phase(qiskit_u, braket_u, atol=1e-7)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for managing qBraid-specific runtime options.

"""

import copy
from dataclasses import dataclass
from dataclasses import field as dataclass_field
from typing import Any, Callable


@dataclass
class RuntimeOptions:
    """
    Manages qBraid-specific options with controlled defaults and dynamic fields.

    The `RuntimeOptions` class allows you to initialize options with default fields,
    add dynamic fields, and validate the values of specific fields using custom
    validators. It also provides dictionary-like access to option fields.

    Example:

    .. code-block:: python

        >>> from qbraid.runtime.options import RuntimeOptions

        # Initialize options with default and custom fields
        >>> options = RuntimeOptions(check=True, custom_field=42)

        # Access option fields like dictionary
        >>> options["check"]
        True

        # Add dynamic fields
        >>> options["new_field"] = "hello"
        >>> print(options)
        RuntimeOptions(check=True, custom_field=42, new_field='hello')

        # Set a validator for a field
        >>> options.set_validator("check", lambda x: isinstance(x, bool))

        # Attempting to set an invalid value will raise a ValueError
        >>> options.check = False  # Valid
        >>> options.check = "invalid_value"  # Raises ValueError

        # Make a shallow copy of the options
        >>> options_copy = copy.copy(options)
        >>> print(options_copy)
        RuntimeOptions(check=False, custom_field=42, new_field='hello')

    Args:
        kwargs: Keyword arguments representing the initial options to set.
                These will form the default fields of the `Options` instance.

    Attributes:
        _fields (dict): The internal dictionary storing option fields.
        _validators (dict): A dictionary storing field validators.

    """

    _validators: dict[str, Callable[[Any], bool]] = dataclass_field(
        default_factory=dict, repr=False
    )

    def __init__(self, **kwargs: Any):
        reserved_keywords = {"set_validator", "validate_option", "update_options", "get"}

        for key in kwargs:
            if key in reserved_keywords:
                raise ValueError(f"The option name '{key}' is reserved and cannot be used.")
            if key.startswith("__"):
                raise ValueError(f"The option name '{key}' with dunder prefix is not allowed.")

        object.__setattr__(self, "_default_fields", kwargs.copy())
        object.__setattr__(self, "_fields", kwargs.copy())
        object.__setattr__(self, "_validators", {})

    def set_validator(self, option_name: str, validator: Callable[[Any], bool]):
        """Sets a validator for a specific field."""
        if hasattr(self, option_name) or option_name in self._fields:
            value = self.get(option_name)
            existing_validator = self._validators.get(option_name)
            error_prefix = f"Existing value '{value}' for field '{option_name}'"
            error_suffix = (
                f"Please {'update the value or ' if not existing_validator else ''}"
                "delete the field before setting the new validator."
            )
            try:
                is_valid = validator(value)
            except Exception as err:
                raise ValueError(
                    f"{error_prefix} raised an exception against "
                    f"the new validator: {err}. {error_suffix}"
                ) from err
            if not is_valid:
                raise ValueError(
                    f"{error_prefix} is not valid for the new validator. {error_suffix}"
                )
        self._validators[option_name] = validator

    def validate_option(self, option_name: str, value: Any):
        """Validates a field's value using the registered validator, if any.

        Raises:
            ValueError: If the validator function raises an exception or returns False.
        """
        validator = self._validators.get(option_name)
        if validator:
            try:
                is_valid = validator(value)
            except Exception as err:
                raise ValueError(
                    f"Validator for field '{option_name}' raised an exception: {err}"
                ) from err
            if not is_valid:
                raise ValueError(f"Value '{value}' is not valid for field '{option_name}'.")

    def update_options(self, **new_options):
        """Updates multiple options with validation."""
        for option_name, value in new_options.items():
            if hasattr(self, option_name):
                self.validate_option(option_name, value)
                setattr(self, option_name, value)
            else:
                self._fields[option_name] = value

    def get(self, key: str, default: Any = None) -> Any:
        """Return the value for the given key, or the default value if not found."""
        if hasattr(self, key):
            return getattr(self, key)
        return self._fields.get(key, default)

    def __getitem__(self, key: str) -> Any:
        """Allow dictionary-like access."""
        if hasattr(self, key):
            return getattr(self, key)
        return self._fields[key]

    def __setitem__(self, key: str, value: Any):
        """Allow dictionary-like updates."""
        self.update_options(**{key: value})

    def __delitem__(self, key: str):
        """Prevent deletion of default fields but allow deletion of additional fields."""
        if key in self._default_fields:
            raise KeyError(f"Cannot delete default field '{key}'.")
        if key in self._fields:
            del self._fields[key]
        else:
            raise KeyError(f"Field '{key}' not found in options.")

    def __getattr__(self, name: str) -> Any:
        """Access options like attributes."""
        if name in {"_fields", "_default_fields", "_validators"}:
            return self.__dict__.get(name)
        if name in self._fields:
            return self._fields[name]
        raise AttributeError(f"Option '{name}' is not defined.")

    def __setattr__(self, name: str, value: Any):
        """Set options like attributes, ensuring validation if necessary."""
        if name in {"_fields", "_default_fields", "_validators"}:
            super().__setattr__(name, value)
        elif name in self._fields or name in self._default_fields:
            if self._fields.get(name) != value:
                if name in self._validators:
                    self.validate_option(name, value)
                self._fields[name] = value
        else:
            self.update_options(**{name: value})

    @property
    def __dict__(self):
        """Returns the internal fields as a dictionary."""
        return self._fields

    def __iter__(self):
        """Allow iteration over the key-value pairs of the internal fields."""
        return iter(self._fields.items())

    def __len__(self):
        """Return the number of fields."""
        return len(self._fields)

    def __repr__(self):
        """Returns a string representation of the options."""
        options_str = ", ".join(f"{k}={v!r}" for k, v in self._fields.items())
        return f"RuntimeOptions({options_str})"

    def __copy__(self):
        """Create a shallow copy of the RuntimeOptions object."""
        cls = self.__class__
        new_instance = cls(**self._fields)
        new_instance._validators = copy.copy(self._validators)
        return new_instance

    def __eq__(self, other: Any) -> bool:
        """Custom equality check for RuntimeOptions objects."""
        if not isinstance(other, RuntimeOptions):
            return False
        if self._fields != other._fields:
            return False
        if set(self._validators.keys()) != set(other._validators.keys()):
            return False
        return True

    def merge(self, other: "RuntimeOptions", override_validators: bool = True):
        """Merges another RuntimeOptions instance into this one.

        Args:
            other (RuntimeOptions): The RuntimeOptions instance to merge from.
            override_validators (bool): Determines whether validators from `other`
                should override existing validators in `self`.

        Raises:
            ValueError: If any option value is invalid after merging.
        """
        combined_validators = self._validators.copy()
        if override_validators:
            combined_validators.update(other._validators)
        else:
            for key, validator in other._validators.items():
                combined_validators.setdefault(key, validator)

        combined_fields = self._fields.copy()
        combined_fields.update(other.__dict__)

        object.__setattr__(self, "_validators", combined_validators)
        object.__setattr__(self, "_fields", combined_fields)

        if override_validators:
            options_to_validate = combined_fields.items()
        else:
            options_to_validate = ((key, combined_fields[key]) for key in other.__dict__.keys())

        for option_name, value in options_to_validate:
            validator = combined_validators.get(option_name)
            if validator and not validator(value):
                raise ValueError(
                    f"Value '{value}' is not valid for field '{option_name}' after merging."
                )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing one-step functions for converting between supported
quantum software program types.

.. currentmodule:: qbraid.transpiler.conversions

Submodules
-----------

.. autosummary::
   :toctree: ../stubs/

   braket_ahs
   braket
   cirq
   openqasm3
   qasm2
   qasm3
   pennylane
   pyquil
   pytket
   qiskit
   cudaq

"""
import importlib
import inspect

# Dynamically import QPROGRAM_ALIASES when needed
_qbraid = importlib.import_module("qbraid.programs._import")
_registry = importlib.import_module("qbraid.programs.registry")
NATIVE_REGISTRY = getattr(_qbraid, "NATIVE_REGISTRY", {})
QPROGRAM_REGISTRY = getattr(_registry, "QPROGRAM_REGISTRY", {})

# Cache for storing previously seen valid combinations, including reversed pairs
valid_combinations_cache = set()

conversion_functions = []


def _update_registered_conversions() -> None:
    """
    Dynamically update the list of conversion functions based on current
    NATIVE_REGISTRY and QPROGRAM_REGISTRY. Adds valid conversion functions
    to the global namespace and maintains a cache of seen valid combinations.
    """
    conversion_functions.clear()

    # Base path for the sub-modules
    base_path = "qbraid.transpiler.conversions."

    for lib in NATIVE_REGISTRY:
        try:
            # Dynamically import the sub-module
            sub_module = importlib.import_module(base_path + lib)

            # Extract function names from the sub-module
            function_names = [
                name
                for name in dir(sub_module)
                if callable(getattr(sub_module, name))
                and inspect.isfunction(getattr(sub_module, name))
            ]

            # Add functions to the current namespace
            for name in function_names:
                p1, p2 = name.split("_to_")
                # Create tuples for both pair and its reverse
                pair = (p1, p2)
                reverse_pair = (p2, p1)

                # Check if either pair or its reverse has been seen as valid before
                if pair in valid_combinations_cache or reverse_pair in valid_combinations_cache:
                    globals()[name] = getattr(sub_module, name)
                    conversion_functions.append(name)
                    continue

                # Check if both p1 and p2 are in the set
                if (p1 in NATIVE_REGISTRY or p2 in NATIVE_REGISTRY) and (
                    p1 in QPROGRAM_REGISTRY and p2 in QPROGRAM_REGISTRY
                ):
                    globals()[name] = getattr(sub_module, name)
                    conversion_functions.append(name)
                    # Add both the pair and its reverse to the cache
                    valid_combinations_cache.add(pair)
                    valid_combinations_cache.add(reverse_pair)

        except ModuleNotFoundError:
            pass


_update_registered_conversions()



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining PyQuboModel Class

"""
from __future__ import annotations

from pyqubo import Model

from qbraid.programs.exceptions import ProgramTypeError

from ._model import AnnealingProgram, QuboProblem


class PyQuboModel(AnnealingProgram):
    """AnnealingProblem subclass that accepts a cpp_pyqubo.Model."""

    def __init__(self, program: Model):
        super().__init__(program)
        if not isinstance(program, Model):
            raise ProgramTypeError(
                message=f"Expected 'pyqubo.Model' object, got '{type(program)}'."
            )

    def to_problem(self, **kwargs) -> QuboProblem:
        """Converts the cpp_pyqubo.Model to a Problem instance."""
        qubo, _ = self.program.to_qubo(**kwargs)
        coefficients = {}

        for key, value in qubo.items():
            coefficients[key] = value

        return QuboProblem(coefficients)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Configures the logging for the qBraid-SDK.

The log level is set based on the `LOG_LEVEL` environment variable, which can be a string
(e.g., "DEBUG", "INFO") or an integer (e.g., 10 for DEBUG). If an invalid value is provided,
the log level defaults to `WARNING`, and a warning is logged.

Format: `%(levelname)s - %(message)s`

Usage:
    from qbraid._logging import logger

Environment Variables:
    LOG_LEVEL: The desired log level (default: WARNING).
"""

import logging
import os
from typing import Literal

VALID_LOG_LEVELS = {logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL}
DEFAULT_LOG_LEVEL = logging.WARNING

LOG_LEVEL_ENV = os.getenv("LOG_LEVEL", "")


def parse_log_level(input_level: str | int | None) -> Literal[10, 20, 30, 40, 50]:
    """Parse the log level from a string or integer and return the corresponding
    integer value. Returns `logging.WARNING` (30) if input is invalid."""
    if input_level is None:
        return DEFAULT_LOG_LEVEL

    if isinstance(input_level, int) or (isinstance(input_level, str) and input_level.isdigit()):
        log_level = int(input_level)
        if log_level in VALID_LOG_LEVELS:
            return log_level

        logging.warning(
            "Invalid log level (int) in LOG_LEVEL: %s. Falling back to WARNING.", log_level
        )
        return DEFAULT_LOG_LEVEL

    log_level_str = input_level.upper()
    log_level = getattr(logging, log_level_str, None)

    if log_level in VALID_LOG_LEVELS:
        return log_level

    if input_level:
        logging.warning(
            "Invalid log level (str) in LOG_LEVEL: %s. Falling back to WARNING.", input_level
        )

    return DEFAULT_LOG_LEVEL


if LOG_LEVEL_ENV:  # pragma: no cover
    logging.basicConfig(format="%(levelname)s - %(message)s", level=parse_log_level(LOG_LEVEL_ENV))

logger = logging.getLogger(__name__)


__all__ = ["logger", "DEFAULT_LOG_LEVEL", "VALID_LOG_LEVELS"]



# Copyright (C) 2024 qBraid
# Copyright (C) 2020 The Cirq Developers
#
# This file is part of the qBraid-SDK.
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. This specific file, adapted from Cirq, is dual-licensed under both the
# Apache License, Version 2.0, and the GPL v3. You may not use this file except in
# compliance with the applicable license. You may obtain a copy of the Apache License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file includes code adapted from Cirq (https://github.com/quantumlib/Cirq)
# with modifications by qBraid. The original copyright notice is included above.
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header

"""
Module for conversions from Quil to Cirq.

"""

from typing import Callable, Union, cast

import numpy as np
from cirq import Circuit, LineQubit
from cirq.ops import (
    CCNOT,
    CNOT,
    CSWAP,
    CZ,
    ISWAP,
    SWAP,
    CZPowGate,
    Gate,
    H,
    I,
    ISwapPowGate,
    MatrixGate,
    MeasurementGate,
    S,
    T,
    TwoQubitDiagonalGate,
    X,
    Y,
    Z,
    ZPowGate,
    rx,
    ry,
    rz,
)
from pyquil import Program
from pyquil.quilbase import Declare
from pyquil.quilbase import Gate as PyQuilGate
from pyquil.quilbase import Measurement as PyQuilMeasurement
from pyquil.quilbase import Pragma, Reset, ResetQubit


class UndefinedQuilGate(Exception):
    """Raised when a Quil gate is not defined."""


class UnsupportedQuilInstruction(Exception):
    """Raised when a Quil instruction is not supported."""


def cphase(param: float) -> CZPowGate:
    """Returns a controlled-phase gate as a Cirq CZPowGate with exponent
    determined by the input param. The angle parameter of pyQuil's CPHASE
    gate and the exponent of Cirq's CZPowGate differ by a factor of pi.

    Args:
        param: Gate parameter (in radians).

    Returns:
        A CZPowGate equivalent to a CPHASE gate of given angle.
    """
    return CZPowGate(exponent=param / np.pi)


def cphase00(phi: float) -> TwoQubitDiagonalGate:
    """Returns a Cirq TwoQubitDiagonalGate for pyQuil's CPHASE00 gate.

    In pyQuil, CPHASE00(phi) = diag([exp(1j * phi), 1, 1, 1]), and in Cirq,
    a TwoQubitDiagonalGate is specified by its diagonal in radians, which
    would be [phi, 0, 0, 0].

    Args:
        phi: Gate parameter (in radians).

    Returns:
        A TwoQubitDiagonalGate equivalent to a CPHASE00 gate of given angle.
    """
    return TwoQubitDiagonalGate([phi, 0, 0, 0])


def cphase01(phi: float) -> TwoQubitDiagonalGate:
    """Returns a Cirq TwoQubitDiagonalGate for pyQuil's CPHASE01 gate.

    In pyQuil, CPHASE01(phi) = diag(1, [exp(1j * phi), 1, 1]), and in Cirq,
    a TwoQubitDiagonalGate is specified by its diagonal in radians, which
    would be [0, phi, 0, 0].

    Args:
        phi: Gate parameter (in radians).

    Returns:
        A TwoQubitDiagonalGate equivalent to a CPHASE01 gate of given angle.
    """
    return TwoQubitDiagonalGate([0, phi, 0, 0])


def cphase10(phi: float) -> TwoQubitDiagonalGate:
    """Returns a Cirq TwoQubitDiagonalGate for pyQuil's CPHASE10 gate.

    In pyQuil, CPHASE10(phi) = diag(1, 1, [exp(1j * phi), 1]), and in Cirq,
    a TwoQubitDiagonalGate is specified by its diagonal in radians, which
    would be [0, 0, phi, 0].

    Args:
        phi: Gate parameter (in radians).

    Returns:
        A TwoQubitDiagonalGate equivalent to a CPHASE10 gate of given angle.
    """
    return TwoQubitDiagonalGate([0, 0, phi, 0])


def phase(param: float) -> ZPowGate:
    """Returns a single-qubit phase gate as a Cirq ZPowGate with exponent
    determined by the input param. The angle parameter of pyQuil's PHASE
    gate and the exponent of Cirq's ZPowGate differ by a factor of pi.

    Args:
        param: Gate parameter (in radians).

    Returns:
        A ZPowGate equivalent to a PHASE gate of given angle.
    """
    return ZPowGate(exponent=param / np.pi)


def pswap(phi: float) -> MatrixGate:
    """Returns a Cirq MatrixGate for pyQuil's PSWAP gate.

    Args:
        phi: Gate parameter (in radians).

    Returns:
        A MatrixGate equivalent to a PSWAP gate of given angle.
    """
    # fmt: off
    pswap_matrix = np.array(
        [
            [1, 0, 0, 0],
            [0, 0, np.exp(1j * phi), 0],
            [0, np.exp(1j * phi), 0, 0],
            [0, 0, 0, 1]
        ],
        dtype=complex,
    )
    # fmt: on
    return MatrixGate(pswap_matrix)


def xy(param: float) -> ISwapPowGate:
    """Returns an ISWAP-family gate as a Cirq ISwapPowGate with exponent
    determined by the input param. The angle parameter of pyQuil's XY gate
    and the exponent of Cirq's ISwapPowGate differ by a factor of pi.

    Args:
        param: Gate parameter (in radians).

    Returns:
        An ISwapPowGate equivalent to an XY gate of given angle.
    """
    return ISwapPowGate(exponent=param / np.pi)


PRAGMA_ERROR = """
Please remove PRAGMAs from your Quil program.
If you would like to add noise, do so after conversion.
"""

RESET_ERROR = """
Please remove RESETs from your Quil program.
RESET directives have special meaning on QCS, to enable active reset.
"""

# Parameterized gates map to functions that produce Gate constructors.
SUPPORTED_GATES: dict[str, Union[Gate, Callable[..., Gate]]] = {
    "CCNOT": CCNOT,
    "CNOT": CNOT,
    "CSWAP": CSWAP,
    "CPHASE": cphase,
    "CPHASE00": cphase00,
    "CPHASE01": cphase01,
    "CPHASE10": cphase10,
    "CZ": CZ,
    "PHASE": phase,
    "H": H,
    "I": I,
    "ISWAP": ISWAP,
    "PSWAP": pswap,
    "RX": rx,
    "RY": ry,
    "RZ": rz,
    "S": S,
    "SWAP": SWAP,
    "T": T,
    "X": X,
    "Y": Y,
    "Z": Z,
    "XY": xy,
}


def parse_defgates(quil_str: str) -> dict[str, np.ndarray]:
    """
    Parses non-parameterized DEFGATE definitions from a Quil program string.

    This function scans through the given Quil program string and extracts
    the definitions of custom gates defined using the DEFGATE directive.
    It only supports non-parameterized gate definitions. If a parameterized
    gate definition is encountered, the function raises a ValueError.

    Args:
        quil_str (str): A string representation of a Quil program containing
                        one or more DEFGATE definitions.

    Returns:
        dict: A dictionary where each key is the name of a custom gate and
              each value is a numpy array representing the gate's unitary matrix.

    Raises:
        UnsupportedQuilInstruction: If a parameterized DEFGATE definition is encountered.
    """
    defgates = {}
    lines = quil_str.split("\n")
    i = 0
    while i < len(lines):
        line = lines[i].strip()
        if line.startswith("DEFGATE"):
            gate_name = line.split()[1]
            # Check for parameters and raise an exception if found
            if "(" in gate_name or ")" in gate_name:
                raise UnsupportedQuilInstruction("Parameterized DEFGATEs are not supported.")

            gate_name = gate_name.split(":")[0]  # Extract the gate name
            matrix_lines = []
            i += 1
            while i < len(lines) and lines[i].strip() and not lines[i].startswith("DEFGATE"):
                # Split the line by comma and whitespace, then strip each element
                matrix_elements = [elem.strip() for elem in lines[i].split(",")]
                matrix_lines.append(matrix_elements)
                i += 1
            # Convert the string elements to floats
            matrix = np.array([[float(element) for element in row] for row in matrix_lines])
            defgates[gate_name] = matrix
        else:
            i += 1
    return defgates


def circuit_from_quil(quil: str) -> Circuit:
    """Convert a Quil program to a Cirq Circuit.

    Args:
        quil: The Quil program to convert.

    Returns:
        A Cirq Circuit generated from the Quil program.

    Raises:
        UnsupportedQuilInstruction: Cirq does not support the specified Quil instruction.
        UndefinedQuilGate: Cirq does not support the specified Quil gate.

    References:
        https://github.com/rigetti/pyquil
    """
    circuit = Circuit()
    defined_gates = SUPPORTED_GATES.copy()
    instructions = Program(quil).instructions

    defgates = parse_defgates(quil)

    # Add the parsed DefGates to defined_gates
    for gate_name, matrix in defgates.items():
        defined_gates[gate_name] = MatrixGate(matrix)

    for inst in instructions:
        # Pass when encountering a DECLARE.
        if isinstance(inst, Declare):
            pass

        # Convert pyQuil gates to Cirq operations.
        elif isinstance(inst, PyQuilGate):
            quil_gate_name = inst.name
            quil_gate_params = inst.params
            line_qubits = list(LineQubit(q.index) for q in inst.qubits)
            if quil_gate_name not in defined_gates:
                raise UndefinedQuilGate(f"Quil gate {quil_gate_name} not supported in Cirq.")
            cirq_gate_fn = defined_gates[quil_gate_name]
            if quil_gate_params:
                circuit += cast(Callable[..., Gate], cirq_gate_fn)(*quil_gate_params)(*line_qubits)
            else:
                circuit += cirq_gate_fn(*line_qubits)

        # Convert pyQuil MEASURE operations to Cirq MeasurementGate objects.
        elif isinstance(inst, PyQuilMeasurement):
            line_qubit = LineQubit(inst.qubit.index)
            if inst.classical_reg is None:
                raise UnsupportedQuilInstruction(
                    f"Quil measurement {inst} without classical register "
                    f"not currently supported in Cirq."
                )
            quil_memory_reference = inst.classical_reg.out()
            circuit += MeasurementGate(1, key=quil_memory_reference)(line_qubit)

        # Raise a targeted error when encountering a PRAGMA.
        elif isinstance(inst, Pragma):
            raise UnsupportedQuilInstruction(PRAGMA_ERROR)

        # Raise a targeted error when encountering a RESET.
        elif isinstance(inst, (Reset, ResetQubit)):
            raise UnsupportedQuilInstruction(RESET_ERROR)

        # Raise a general error when encountering an unconsidered type.
        else:
            raise UnsupportedQuilInstruction(
                f"Quil instruction {inst} of type {type(inst)} not currently supported in Cirq."
            )

    return circuit



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Benchmarking tests for Cirq conversions

"""
import importlib.util
import string

import cirq
import numpy as np
import pytest
import scipy

from qbraid.interface import circuits_allclose
from qbraid.transpiler import ConversionGraph, transpile


def generate_params(varnames, seed=0):
    """Generate random parameters to help construct Cirq test gates"""
    np.random.seed(seed)
    params = {}
    # Generating angles
    for ra in ["rads", "theta", "phi"]:
        if ra in varnames:
            params[ra] = np.random.rand() * 2 * np.pi
    # Generating probabilities
    for prob in ["gamma", "p", "probability"]:
        if prob in varnames:
            params[prob] = np.random.rand()

    # exponents
    for exp in ["exponent", "x_exponent", "z_exponent", "phase_exponent", "axis_phase_exponent"]:
        if exp in varnames:
            params[exp] = np.random.rand() * 10

    if "num_qubits" in varnames:
        params["num_qubits"] = np.random.randint(1, 7)

    if "sub_gate" in varnames:
        params["sub_gate"] = np.random.choice([cirq.X, cirq.Y, cirq.Z, cirq.S, cirq.T])

    if "matrix" in varnames:
        n = np.random.randint(1, 4)
        params["matrix"] = scipy.stats.unitary_group.rvs(2**n)

    return params


def get_cirq_gates():
    """Construct a dictionary of all Cirq gates and assign random parameters as applicable"""
    qubits = cirq.LineQubit.range(7)

    cirq_gate_dict = {
        attr: None
        for attr in dir(cirq.ops)
        if attr[0] in string.ascii_uppercase
        if (
            isinstance(
                getattr(cirq.ops, attr), (cirq.Gate, cirq.value.abc_alt.ABCMetaImplementAnyOneOf)
            )
        )
    }
    for gate in cirq_gate_dict:
        try:
            params = generate_params(getattr(cirq.ops, gate).__init__.__code__.co_varnames)
        except Exception:  # pylint: disable=broad-exception-caught
            continue

        if not isinstance(getattr(cirq.ops, gate), cirq.Gate):
            try:
                cirq_gate_dict[gate] = getattr(cirq.ops, gate)(**params)
            except Exception:  # pylint: disable=broad-exception-caught
                continue

    for gate in cirq_gate_dict:
        if cirq_gate_dict.get(gate):
            cirq_gate_dict[gate] = cirq_gate_dict[gate](
                *qubits[: cirq_gate_dict[gate].num_qubits()]
            )
        else:
            try:
                cirq_gate_dict[gate] = getattr(cirq.ops, gate)(
                    *qubits[: getattr(cirq.ops, gate).num_qubits()]
                )
            except Exception:  # pylint: disable=broad-exception-caught
                continue

    return {k: v for k, v in cirq_gate_dict.items() if v is not None}


def is_package_installed(package_name: str) -> bool:
    """Check if a package is installed."""
    return importlib.util.find_spec(package_name) is not None


ALL_TARGETS = [("braket", 0.85), ("pyquil", 0.74), ("pytket", 0.87), ("qiskit", 0.87)]
AVAILABLE_TARGETS = [(name, version) for name, version in ALL_TARGETS if is_package_installed(name)]

cirq_gates = get_cirq_gates()

graph = ConversionGraph(require_native=True)


def convert_from_cirq_to_x(target, gate_name):
    """Construct a Cirq circuit with the given gate, transpile it to
    target program type, and check equivalence.
    """
    gate = cirq_gates[gate_name]
    source_circuit = cirq.Circuit()
    source_circuit.append(gate)
    target_circuit = transpile(source_circuit, target, conversion_graph=graph)
    assert circuits_allclose(source_circuit, target_circuit, strict_gphase=False)


@pytest.mark.parametrize(("target", "baseline"), AVAILABLE_TARGETS)
def test_cirq_coverage(target, baseline):
    """Test converting Cirq circuits to supported target program type over
    all Cirq gates and check against baseline expecte accuracy.
    """
    ACCURACY_BASELINE = baseline
    ALLOWANCE = 0.01
    failures = {}
    for gate_name in cirq_gates:
        try:
            convert_from_cirq_to_x(target, gate_name)
        except Exception as e:  # pylint: disable=broad-exception-caught
            failures[f"{target}-{gate_name}"] = e

    total_tests = len(cirq_gates)
    nb_fails = len(failures)
    nb_passes = total_tests - nb_fails
    accuracy = float(nb_passes) / float(total_tests)

    assert accuracy >= ACCURACY_BASELINE - ALLOWANCE, (
        f"The coverage threshold was not met. {nb_fails}/{total_tests} tests failed "
        f"({nb_fails / (total_tests):.2%}) and {nb_passes}/{total_tests} passed "
        f"(expected >= {ACCURACY_BASELINE}).\nFailures: {failures.keys()}\n\n"
    )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining AnalogHamiltonianProgram Class

"""
from __future__ import annotations

import json
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from qbraid.programs.program import QuantumProgram

if TYPE_CHECKING:
    import qbraid.runtime


class AnalogHamiltonianProgram(QuantumProgram, ABC):
    """Abstract class for qbraid program wrapper objects."""

    @property
    def hamiltonian(self) -> dict:
        """Return the Hamiltonian of the program."""
        return self.to_dict()["hamiltonian"]

    @property
    def register(self) -> dict:
        """AtomArrangement: The initial atom arrangement for the simulation."""
        return self.to_dict()["register"]

    @property
    def num_atoms(self) -> int:
        """Return the number of atoms in the program."""
        return len(self.register["sites"])

    @property
    def num_qubits(self) -> int:
        """Number of qubits needed by a quantum device to execute this program."""
        return self.num_atoms

    def transform(self, device: qbraid.runtime.QuantumDevice) -> None:
        """Transform program to according to device target profile."""
        return None

    @abstractmethod
    def to_dict(self) -> dict:
        """Return the dictionary representation of the program."""

    def serialize(self) -> dict[str, str]:
        """Return the program in a format suitable for submission to the qBraid API."""
        return {"ahs": json.dumps(self, cls=AHSEncoder)}

    def __eq__(self, other: Any) -> bool:
        """Check if two programs are equal."""
        if not isinstance(other, self.__class__):
            return False
        return self.to_dict() == other.to_dict()


class AHSEncoder(json.JSONEncoder):
    """Class for encoding AnalogHamiltonianProgram objects to dictionaries.

    Example:

    .. code-block:: python

        >>> import json
        >>> from qbraid.programs.ahs import AHSEncoder
        >>> json_str = json.dumps(ahs_program, cls=AHSEncoder)

    """

    def default(self, o: Any) -> Any:
        if isinstance(o, AnalogHamiltonianProgram):
            return o.to_dict()
        return super().default(o)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for the typer module

"""

from unittest.mock import patch

import pytest
from pyqasm.exceptions import QasmParsingError

from qbraid.programs.typer import (
    BaseQasmInstanceMeta,
    IonQDict,
    IonQDictInstanceMeta,
    ProgramValidationError,
    Qasm2KirinString,
    Qasm2String,
    Qasm2StringType,
    Qasm3String,
    Qasm3StringType,
    QasmStringType,
    QuboCoefficientsDict,
    get_qasm_type_alias,
)

valid_qasm2_string = """
OPENQASM 2.0;
include "qelib1.inc";
qreg q[2];
swap q[0],q[1];
"""

valid_qasm3_string = """
OPENQASM 3.0;
include "stdgates.inc";
qubit[1] q;
z q[0];
"""

valid_qasm2_kirin_string = """
KIRIN {cf,func,py.ilist,qasm2.core,qasm2.expr,qasm2.glob,qasm2.indexing,qasm2.inline,qasm2.noise,qasm2.parallel,qasm2.uop};
include "qelib1.inc";
qreg qreg[4];
CX qreg[0], qreg[1];
reset qreg[0];
parallel.CZ {
  qreg[0], qreg[2];
  qreg[1], qreg[3];
}
"""

invalid_qasm_string = "Some random string"


@pytest.mark.parametrize(
    "cls, string",
    [
        (Qasm2StringType, valid_qasm2_string),
        (Qasm3StringType, valid_qasm3_string),
    ],
)
def test_qasm_string_valid(cls, string):
    """Test that the Qasm2StringType and Qasm3StringType
    classes can be instantiated with valid strings."""
    assert isinstance(cls(string), cls)


@pytest.mark.parametrize(
    "cls, string, error",
    [
        (Qasm2StringType, valid_qasm3_string, ValueError),
        (Qasm3StringType, invalid_qasm_string, QasmParsingError),
    ],
)
def test_qasm_string_invalid(cls, string, error):
    """Test that the Qasm2StringType and Qasm3StringType classes raise an error
    when instantiated with invalid strings."""
    with pytest.raises(error):
        cls(string)


@pytest.mark.parametrize(
    "meta, string",
    [
        (Qasm2String, valid_qasm2_string),
        (Qasm3String, valid_qasm3_string),
        (Qasm2KirinString, valid_qasm2_kirin_string),
    ],
)
def test_isinstance_checks_valid(meta, string):
    """Test that the isinstance function correctly identifies valid OpenQASM strings."""
    assert isinstance(string, meta)


def test_get_qasm_type_alias_kirin():
    """Test that get_qasm_type_alias returns the correct alias for KIRIN strings."""
    alias = get_qasm_type_alias(valid_qasm2_kirin_string)
    assert alias == "qasm2_kirin"  # pylint: disable=comparison-with-callable


@pytest.mark.parametrize(
    "meta, version, string",
    [
        (Qasm2String, 2, valid_qasm3_string),
        (Qasm3String, 3, invalid_qasm_string),
    ],
)
def test_isinstance_checks_invalid(meta, version, string):
    """Test that the isinstance function correctly identifies invalid OpenQASM strings."""
    with patch("pyqasm.analyzer.Qasm3Analyzer.extract_qasm_version", return_value=version + 1):
        assert not isinstance(string, meta)


def test_qasm_string_type_error():
    """Test that the QasmString class raises a TypeError when instantiated with an invalid type."""
    with pytest.raises(TypeError, match="OpenQASM strings must be initialized with a string."):
        QasmStringType(123)


def test_base_qasm_instance_meta():
    """Test that the BaseQasmInstanceMeta metaclass does not
    recognize dictionaries as instances of its subclasses."""

    class MockQasmInstance(metaclass=BaseQasmInstanceMeta):
        """Mock class for testing the BaseQasmInstanceMeta metaclass."""

    assert not isinstance({}, MockQasmInstance)


def test_base_qasm_instance_meta_alias():
    """Test that the BaseQasmInstanceMeta metaclass has the correct alias."""

    class FakeQasmInstance(metaclass=BaseQasmInstanceMeta):
        """Fake instance class for testing"""

    assert FakeQasmInstance.__alias__ is None


def test_not_qasm_string():
    """Test that the isinstance function returns False when given a non-QasmString object."""
    assert not isinstance("Not a QasmString", Qasm2String)


@pytest.mark.parametrize(
    "qasm_instance_class, version",
    [
        (Qasm2String, 2),
        (Qasm3String, 3),
    ],
)
def test_qasm_instance_properties(qasm_instance_class, version):
    """Test that the QasmInstance class has the correct properties for different versions."""
    assert qasm_instance_class.version == version
    assert qasm_instance_class.__alias__ == f"qasm{version}"
    assert qasm_instance_class.__bound__ is str


def test_ionq_dict_instance_meta_alias():
    """Test that __alias__ property returns the correct alias."""
    assert IonQDict.__alias__ == "ionq"  # pylint: disable=comparison-with-callable


def test_ionq_dict_instance_meta_bound():
    """Test that __bound__ property returns dict."""
    assert IonQDict.__bound__ is dict  # pylint: disable=comparison-with-callable


def test_ionq_isinstance_valid_instance():
    """Test that the isinstance function correctly identifies valid IonQDict instances."""
    circuit = {
        "qubits": 3,
        "circuit": [
            {"gate": "h", "target": 0},
            {"gate": "cnot", "control": 0, "target": 1},
            {"gate": "cnot", "control": 0, "target": 2},
        ],
    }
    assert isinstance(circuit, IonQDict)


@pytest.mark.parametrize(
    "invalid_instance",
    [
        {"qubits": 2, "circuit": [{"gate": 123, "target": 0}]},
        {"qubits": 2, "circuit": [{"gate": "rx", "rotation": "invalid", "target": 0}]},
        {"qubits": 2, "circuit": [{"gate": "cx", "target": "invalid"}]},
        {"qubits": 2, "circuit": [{"gate": "cx", "target": 0, "control": "invalid"}]},
        {"qubits": "invalid", "circuit": [{"gate": "cx", "target": 0}]},
        {"qubits": 2, "circuit": "invalid"},
        {"qubits": 2, "circuit": [{"gate": "cx", "target": 0, "control": 1}, 42]},
    ],
)
def test_ionq_instancecheck_invalid_instances(invalid_instance):
    """Test various invalid instances of IonQDictInstanceMeta."""
    assert not isinstance(invalid_instance, IonQDict)


@pytest.mark.parametrize(
    "single, multiple, field_name",
    [
        (1, None, "target"),
        (None, [1, 2], "control"),
        (1, None, "target"),
        (None, [1, 2, 3], "target"),
    ],
)
def test_ionq_dict_instance_meta_validate_field_valid_cases(single, multiple, field_name):
    """Test _validate_field with valid single and multiple fields."""
    IonQDictInstanceMeta._validate_field(single=single, multiple=multiple, field_name=field_name)


@pytest.mark.parametrize(
    "single, multiple, field_name",
    [
        (1, [2, 3], "target"),
        ("invalid", None, "target"),
        (None, "invalid", "target"),
        (None, [1, "invalid", 3], "target"),
    ],
)
def test_ionq_dict_instance_meta_validate_field_invalid_cases(single, multiple, field_name):
    """Test _validate_field raises ProgramValidationError for invalid input cases."""
    with pytest.raises(ProgramValidationError):
        IonQDictInstanceMeta._validate_field(single, multiple, field_name)


def test_qubo_coefficients_dict_valid():
    """Test with a valid QUBO coefficients dictionary."""
    valid_dict = {
        ("s1", "s1"): -160.0,
        ("s4", "s2"): 16.0,
        ("s3", "s1"): 224.0,
        ("s2", "s2"): -96.0,
        ("s4", "s1"): 32.0,
        ("s1", "s2"): 64.0,
        ("s3", "s2"): 112.0,
        ("s3", "s3"): -196.0,
        ("s4", "s4"): -52.0,
        ("s4", "s3"): 56.0,
    }
    assert isinstance(valid_dict, QuboCoefficientsDict)


def test_qubo_coefficients_dict_invalid_key_type():
    """Test with an invalid key type (non-tuple key)."""
    invalid_dict = {"s1": -160.0}  # Invalid because key is not a tuple
    assert not isinstance(invalid_dict, QuboCoefficientsDict)


def test_qubo_coefficients_dict_invalid_key_tuple_elements():
    """Test with a tuple key containing non-string elements."""
    invalid_dict = {
        (1, "s2"): 32.0,  # Invalid because one element in the tuple is not a string
        ("s3", 3): 64.0,  # Invalid because one element in the tuple is not a string
    }
    assert not isinstance(invalid_dict, QuboCoefficientsDict)


def test_qubo_coefficients_dict_invalid_key_tuple_length():
    """Test with a tuple key of incorrect length."""
    invalid_dict = {
        ("s1",): -160.0,  # Invalid because tuple length is 1
        ("s2", "s3", "s4"): 96.0,  # Invalid because tuple length is 3
    }
    assert not isinstance(invalid_dict, QuboCoefficientsDict)


def test_qubo_coefficients_dict_invalid_value_type():
    """Test with a value that is not a float."""
    invalid_dict = {
        ("s1", "s2"): 64.0,
        ("s3", "s3"): "string",  # Invalid because the value is not a float
    }
    assert not isinstance(invalid_dict, QuboCoefficientsDict)


def test_qubo_coefficients_dict_empty():
    """Test with an empty dictionary, which should be invalid."""
    empty_dict = {}
    assert not isinstance(empty_dict, QuboCoefficientsDict)


def test_qubo_coefficients_dict_nested_invalid():
    """Test with a nested dictionary, which should be invalid."""
    nested_invalid_dict = {
        ("s1", "s1"): -160.0,
        ("s2", "s2"): {"nested_key": 32.0},  # Invalid because the value is a dict, not a float
    }
    assert not isinstance(nested_invalid_dict, QuboCoefficientsDict)


def test_qubo_coefficients_dict_non_dict_type_invalid():
    """Test with a non-dictionary type, which should be invalid."""
    non_dict = "string"
    assert not isinstance(non_dict, QuboCoefficientsDict)


def test_qubo_coefficients_dict_instance_meta_alias():
    """Test that __alias__ property returns the correct alias."""
    assert QuboCoefficientsDict.__alias__ == "qubo"  # pylint: disable=comparison-with-callable


def test_qubo_coefficients_dictt_instance_meta_bound():
    """Test that __bound__ property returns dict."""
    assert QuboCoefficientsDict.__bound__ is dict  # pylint: disable=comparison-with-callable



# Copyright (C) 2024 qBraid
# Copyright (C) Unitary Fund
#
# This file is part of the qBraid-SDK.
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# This file includes code adapted from Mitiq (https://github.com/unitaryfund/mitiq)
# with modifications by qBraid. The original copyright notice is included above.
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header

"""
Unit tests for converting Cirq circuits to Braket circuits

"""
from unittest.mock import patch

import cirq
import numpy as np
import pytest
from braket.circuits import Gate, Instruction, QubitSet
from braket.circuits import noises as braket_noise_gate
from cirq import Circuit, LineQubit, ops, testing
from cirq_ionq.ionq_native_gates import GPI2Gate, GPIGate, MSGate

from qbraid.interface import circuits_allclose
from qbraid.interface.random import random_unitary_matrix
from qbraid.transpiler.conversions.cirq import cirq_to_braket
from qbraid.transpiler.conversions.cirq.braket_custom import C
from qbraid.transpiler.conversions.cirq.cirq_to_braket import (
    _to_one_qubit_braket_instruction,
    _to_two_qubit_braket_instruction,
)
from qbraid.transpiler.exceptions import ProgramConversionError


def alpha_pump(sys, env):
    """Alpha pump"""
    yield cirq.CNOT(sys[0], env[0])
    yield cirq.H(sys[1])
    yield cirq.X(env[1])


def beta_pump(sys, env):
    """Beta pump"""
    yield cirq.CNOT(sys[1], sys[0])
    yield cirq.X(env[0])


@pytest.fixture
def batch_circuit():
    """Return a batch alpha pump, beta bump circuit"""
    sys_qubits = [cirq.LineQubit(i) for i in range(0, 6)]
    env_qubits = [cirq.LineQubit(i) for i in range(6, 11)]

    circuit = cirq.Circuit()
    circuit += alpha_pump(sys_qubits[0:2], env_qubits[0:2])
    circuit += alpha_pump(sys_qubits[2:4], env_qubits[2:4])
    circuit += beta_pump(sys_qubits[4:6], env_qubits[4:5])
    circuit.append(cirq.measure(*circuit.all_qubits()))
    return circuit


# pylint: disable-next=redefined-outer-name
def test_ommit_measurement_gate(batch_circuit):
    """Test that cirq.Measurementgate is skipped during Braket conversion"""
    braket_circuit = cirq_to_braket(batch_circuit)
    assert circuits_allclose(batch_circuit, braket_circuit)


@pytest.mark.parametrize("qreg", (LineQubit.range(2), [LineQubit(1), LineQubit(6)]))
def test_to_braket_bell_circuit(qreg):
    """Test converting bell circuit"""
    cirq_circuit = Circuit(ops.H(qreg[0]), ops.CNOT(*qreg))
    braket_circuit = cirq_to_braket(cirq_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)


def test_to_braket_non_parameterized_single_qubit_gates():
    """Test converting circuit containing non-parameterized single-qubits gates"""
    qreg = LineQubit.range(4)
    cirq_circuit = Circuit(
        ops.I(qreg[0]),
        ops.X(qreg[1]),
        ops.Y(qreg[2]),
        ops.Z(qreg[3]),
        ops.H(qreg[0]),
        ops.S(qreg[1]),
        ops.S(qreg[2]) ** -1,
        ops.T(qreg[3]),
        ops.T(qreg[0]) ** -1,
        ops.X(qreg[1]) ** 0.5,
        ops.X(qreg[2]) ** -0.5,
    )
    braket_circuit = cirq_to_braket(cirq_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("qubit_index", (0, 3))
def test_to_braket_parameterized_single_qubit_gates(qubit_index):
    """Testing converting circuit containing parameterized single-qubit gates"""
    qubit = LineQubit(qubit_index)
    angles = np.random.RandomState(11).random(4)  # pylint: disable=no-member
    cirq_circuit = Circuit(
        ops.rx(angles[0]).on(qubit),
        ops.ry(angles[1]).on(qubit),
        ops.rz(angles[2]).on(qubit),
        ops.Z.on(qubit) ** (angles[3] / np.pi),
    )
    braket_circuit = cirq_to_braket(cirq_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)


def test_to_braket_non_parameterized_two_qubit_gates():
    """Test converting circuit containing non-parameterized two-qubit gates."""
    qreg = LineQubit.range(2, 5)
    cirq_circuit = Circuit(
        ops.CNOT(*qreg[:2]),
        ops.SWAP(*qreg[1:]),
        ops.ISWAP(*qreg[:2]),
        ops.CZ(*qreg[1:]),
        ops.ControlledGate(ops.Y).on(*qreg[:2]),
    )
    braket_circuit = cirq_to_braket(cirq_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)


def test_to_braket_three_qubit_gates():
    """Testing converting circuits containing three-qubit gates"""
    qreg = LineQubit.range(1, 4)
    cirq_circuit = Circuit(ops.TOFFOLI(*qreg), ops.FREDKIN(*qreg))
    braket_circuit = cirq_to_braket(cirq_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)


def test_to_braket_common_one_qubit_gates():
    """These gates should stay the same (i.e., not get decomposed) when
    converting Cirq -> Braket.
    """
    rots = [ops.rx, ops.ry, ops.rz]
    angles = [1 / 5, 3 / 5, -4 / 5]
    qubit = LineQubit(0)
    cirq_circuit = Circuit(
        # Paulis.
        ops.X(qubit),
        ops.Y(qubit),
        ops.Z(qubit),
        # Single-qubit rotations.
        [rot(angle).on(qubit) for rot, angle in zip(rots, angles)],
        # Rz alter egos.
        ops.T(qubit),
        ops.T(qubit) ** -1,
        ops.S(qubit),
        ops.S(qubit) ** -1,
        # Rx alter egos.
        ops.X(qubit) ** 0.5,
        ops.X(qubit) ** -0.5,
    )

    braket_circuit = cirq_to_braket(cirq_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("uncommon_gate", [ops.HPowGate(exponent=-1 / 14)])
def test_to_braket_uncommon_one_qubit_gates(uncommon_gate):
    """These gates get decomposed when converting Cirq -> Braket, but
    the unitaries should be equal up to global phase.
    """
    cirq_circuit = Circuit(uncommon_gate.on(LineQubit(0)))
    braket_circuit = cirq_to_braket(cirq_circuit)
    assert circuits_allclose(
        cirq_circuit,
        braket_circuit,
        atol=1e-7,
    )


@pytest.mark.parametrize(
    "common_gate",
    [
        ops.CNOT,
        ops.CZ,
        ops.ISWAP,
        ops.XXPowGate(exponent=-0.2),
        ops.YYPowGate(exponent=0.3),
        ops.ZZPowGate(exponent=-0.1),
    ],
)
def test_to_braket_common_two_qubit_gates(common_gate):
    """These gates should stay the same (i.e., not get decomposed) when
    converting Cirq -> Braket.
    """
    cirq_circuit = Circuit(common_gate.on(*LineQubit.range(2)))
    braket_circuit = cirq_to_braket(cirq_circuit)
    if not isinstance(common_gate, (ops.XXPowGate, ops.YYPowGate, ops.ZZPowGate)):
        assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)
    else:
        assert circuits_allclose(braket_circuit, cirq_circuit)


@pytest.mark.parametrize(
    "uncommon_gate",
    [ops.CNotPowGate(exponent=-1 / 17), ops.CZPowGate(exponent=2 / 7)],
)
def test_to_braket_uncommon_two_qubit_gates(uncommon_gate):
    """These gates get decomposed when converting Cirq -> Braket, but
    the unitaries should be equal up to global phase.
    """
    cirq_circuit = Circuit(uncommon_gate.on(*LineQubit.range(2)))
    braket_circuit = cirq_to_braket(cirq_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit)


@pytest.mark.parametrize(
    "common_gate",
    [
        ops.TOFFOLI,
        ops.FREDKIN,
    ],
)
def test_to_braket_common_three_qubit_gates(common_gate):
    """These gates should stay the same (i.e., not get decomposed) when
    converting Cirq -> Braket.
    """
    cirq_circuit = Circuit(common_gate.on(*LineQubit.range(3)))
    braket_circuit = cirq_to_braket(cirq_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("num_qubits", [1, 2, 3, 4, 5])
def test_50_random_circuits(num_qubits):
    """Testing converting 50 random circuits"""
    for i in range(10):
        moments = np.random.randint(1, 6)
        state = num_qubits + i
        cirq_circuit = testing.random_circuit(
            num_qubits, n_moments=moments, op_density=1, random_state=state
        )
        braket_circuit = cirq_to_braket(cirq_circuit)
        if not circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True):
            assert False
    assert True


@pytest.mark.parametrize(
    "noise_gate, target_gate",
    [
        (ops.BitFlipChannel, braket_noise_gate.BitFlip),
        (ops.PhaseFlipChannel, braket_noise_gate.PhaseFlip),
        (ops.DepolarizingChannel, braket_noise_gate.Depolarizing),
    ],
)
def test_to_braket_single_probability_noise_gate(noise_gate, target_gate):
    """Test transpile single arg noise braket"""
    probs = np.random.uniform(low=0, high=0.5)
    cirq_circuit = Circuit(noise_gate(probs).on(*LineQubit.range(1)))
    braket_circuit = cirq_to_braket(cirq_circuit)
    gate = braket_circuit.instructions[0].operator
    assert isinstance(gate, target_gate)
    assert gate.qubit_count == 1
    assert gate.probability == probs


@pytest.mark.parametrize(
    "noise_gate, target_gate",
    [
        (ops.AmplitudeDampingChannel, braket_noise_gate.AmplitudeDamping),
        (ops.PhaseDampingChannel, braket_noise_gate.PhaseDamping),
    ],
)
def test_to_braket_single_gamma_noise_gate(noise_gate, target_gate):
    """Test transpile single arg noise braket"""
    probs = np.random.uniform(low=0, high=0.5)
    cirq_circuit = Circuit(noise_gate(probs).on(*LineQubit.range(1)))
    braket_circuit = cirq_to_braket(cirq_circuit)
    gate = braket_circuit.instructions[0].operator
    assert isinstance(gate, target_gate)
    assert gate.qubit_count == 1
    assert gate.gamma == probs


def test_to_braket_GeneralizedAmplitudeDampingChannel():
    """Test two arg noise gate"""
    probs = np.random.uniform(low=0, high=0.5, size=2)
    cirq_circuit = Circuit(
        ops.GeneralizedAmplitudeDampingChannel(probs[0], probs[1]).on(*LineQubit.range(1))
    )
    braket_circuit = cirq_to_braket(cirq_circuit)
    gate = braket_circuit.instructions[0].operator
    assert isinstance(gate, braket_noise_gate.GeneralizedAmplitudeDamping)
    assert gate.qubit_count == 1
    assert gate.probability == probs[0]
    assert gate.gamma == probs[1]


def test_to_braket_DepolarizingChannel():
    """Test DepolarizingChannel"""
    probs = np.random.uniform(low=0, high=0.5, size=1)
    cirq_circuit = Circuit(ops.DepolarizingChannel(probs[0]).on(*LineQubit.range(1)))
    braket_circuit = cirq_to_braket(cirq_circuit)
    gate = braket_circuit.instructions[0].operator
    assert isinstance(gate, braket_noise_gate.Depolarizing)
    assert gate.qubit_count == 1
    assert gate.probability == probs


def test_to_braket_two_DepolarizingChannel():
    """Test two qubit DepolarizingChannel"""
    probs = np.random.uniform(low=0, high=0.5, size=1)
    cirq_circuit = Circuit(ops.DepolarizingChannel(probs[0], n_qubits=2).on(*LineQubit.range(2)))
    braket_circuit = cirq_to_braket(cirq_circuit)
    gate = braket_circuit.instructions[0].operator
    assert isinstance(gate, braket_noise_gate.TwoQubitDepolarizing)
    assert gate.qubit_count == 2
    assert gate.probability == probs


def test_to_braket_kraus_gates():
    """Test Kraus"""
    K0 = np.sqrt(0.8) * np.eye(4)
    K1 = np.sqrt(0.2) * np.kron(np.array([[0, 1], [1, 0]]), np.array([[0, 1], [1, 0]]))
    cirq_circuit = Circuit(ops.KrausChannel([K0, K1]).on(*LineQubit.range(2)))
    braket_circuit = cirq_to_braket(cirq_circuit)
    gate = braket_circuit.instructions[0].operator
    assert isinstance(gate, braket_noise_gate.Kraus)
    assert gate.qubit_count == 2
    assert np.allclose(gate._matrices, [K0, K1])


def test_braket_unitary_display_name():
    """Test braket unitary gate uses correct display name"""
    unitary = random_unitary_matrix(2)
    braket_circuit = cirq_to_braket(Circuit(ops.MatrixGate(unitary).on(LineQubit(0))))
    assert braket_circuit.instructions[0].operator.ascii_symbols[0] == "U"


def test_ionq_gates():
    """Test IonQ Native gates"""
    q0, q1, q2 = cirq.LineQubit.range(3)
    gpi = GPIGate(phi=0.5).on(q0)
    gpi2 = GPI2Gate(phi=0.5).on(q1)
    ms = MSGate(phi0=0.5, phi1=0.5).on(q1, q2)

    cirq_circuit = cirq.Circuit([gpi, gpi2, ms])
    braket_circuit = cirq_to_braket(cirq_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit)


def test_custom_controlled_three_qubit_gate():
    """Test custom controlled three qubit gate"""
    q0, q1, q2 = cirq.LineQubit.range(3)
    cirq_circuit = cirq.Circuit(ops.ControlledGate(ops.CZ).on(q0, q1, q2))
    braket_circuit = cirq_to_braket(cirq_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit)


def test_three_qubit_matrix_gate():
    """Test three qubit matrix gate"""
    q0, q1, q2 = cirq.LineQubit.range(3)
    matrix = np.eye(8)
    cirq_circuit = cirq.Circuit(ops.MatrixGate(matrix).on(q0, q1, q2))
    braket_circuit = cirq_to_braket(cirq_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit)


def test_three_qubit_error():
    """Test error for three qubit gate"""
    q0, q1, q2 = cirq.LineQubit.range(3)
    matrix = np.eye(8)
    cirq_circuit = cirq.Circuit(ops.MatrixGate(matrix).on(q0, q1, q2))

    with pytest.raises(ProgramConversionError):
        with patch("cirq.protocols.unitary", side_effect=TypeError):
            cirq_to_braket(cirq_circuit)


def test_invalid_gate_instruction():
    """Test error for invalid gate instruction"""
    with pytest.raises(ValueError):
        _to_one_qubit_braket_instruction(1, 1)

    with pytest.raises(ValueError):
        _to_two_qubit_braket_instruction(1, [0, 1])


def test_braket_custom_gate():
    """Test custom gate"""
    custom_gate = C(sub_gate=Gate(1, "a"), targets=QubitSet([0, 1]))
    instr = custom_gate.c(QubitSet([0, 1]), sub_gate=Gate(1, "a"))
    assert isinstance(instr, Instruction)



# Copyright (C) 2024 qBraid
# Copyright (C) Unitary Fund
#
# This file is part of the qBraid-SDK.
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# This file includes code adapted from Mitiq (https://github.com/unitaryfund/mitiq)
# with modifications by qBraid. The original copyright notice is included above.
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header

"""
Module for converting Braket circuits to Cirq circuits

"""
from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Optional, Union

import numpy as np

try:
    from braket.circuits import Circuit as BKCircuit
    from braket.circuits import Instruction as BKInstruction
    from braket.circuits import gates as braket_gates
    from braket.circuits import noises as braket_noise_gate
except ImportError:  # pragma: no cover
    BKCircuit = None
    BKInstruction = None
    braket_gates = None
    braket_noise_gate = None

from cirq import Circuit
from cirq import ops as cirq_ops
from cirq import protocols
from cirq.linalg.decompositions import kak_decomposition

try:
    import cirq_ionq.ionq_native_gates as cirq_ionq_ops
except ImportError:  # pragma: no cover
    cirq_ionq_ops = None

import qbraid.programs.gate_model.cirq
from qbraid.transpiler.annotations import weight
from qbraid.transpiler.exceptions import ProgramConversionError

try:
    from .braket_custom import C as BKControl
except ImportError:  # pragma: no cover
    BKControl = None

if TYPE_CHECKING:
    import braket.circuits


@weight(0.85)
def cirq_to_braket(circuit: Circuit) -> braket.circuits.Circuit:
    """Returns a Braket circuit equivalent to the input Cirq circuit.

    Args:
        circuit: Cirq circuit to convert to a Braket circuit.

    Returns:
        Braket circuit equivalent to the input Cirq circuit.
    """
    cirq_qubits = list(circuit.all_qubits())
    cirq_int_qubits = [
        qbraid.programs.gate_model.cirq.CirqCircuit._int_from_qubit(q) for q in cirq_qubits
    ]
    braket_int_qubits = deepcopy(cirq_int_qubits)
    qubit_mapping = {q: braket_int_qubits[i] for i, q in enumerate(cirq_int_qubits)}
    return BKCircuit(
        _to_braket_instruction(operation, qubit_mapping) for operation in circuit.all_operations()
    )


def _to_braket_instruction(
    operation: cirq_ops.Operation,
    qubit_mapping: dict[int, int],
) -> list[braket.circuits.Instruction]:
    """Converts Cirq operation to equivalent Braket instruction(s).

    Args:
        operation: Cirq operation to convert.
        qubit_mapping: Mappings of input / output qubit indicies

    Raises:
        ProgramConversionError: If the operation cannot be converted to Braket.
    """
    if isinstance(
        operation, (cirq_ops.MeasurementGate, cirq_ops.Operation)
    ) and qbraid.programs.gate_model.cirq.CirqCircuit.is_measurement_gate(operation):
        return []

    nqubits = protocols.num_qubits(operation)
    cirq_qubits = operation.qubits
    cirq_qubits = [
        qbraid.programs.gate_model.cirq.CirqCircuit._int_from_qubit(q) for q in operation.qubits
    ]
    qubits = [qubit_mapping[x] for x in cirq_qubits]

    if nqubits == 1:
        target = qubits[0]
        return _to_one_qubit_braket_instruction(operation, target)

    if nqubits == 2:
        return _to_two_qubit_braket_instruction(operation, qubits)

    if nqubits == 3:
        if operation == cirq_ops.TOFFOLI.on(*operation.qubits):
            return [BKInstruction(braket_gates.CCNot(), qubits)]
        if operation == cirq_ops.FREDKIN.on(*operation.qubits):
            return [BKInstruction(braket_gates.CSwap(), qubits)]
        if isinstance(operation.gate, cirq_ops.ControlledGate):
            sub_gate_instr = _to_two_qubit_braket_instruction(operation.gate.sub_gate, qubits[1:])
            sub_gate = sub_gate_instr[0].operator
            # return [BKInstruction(sub_gate, target=qubits[1:], control=qubits[:1])]
            return [BKInstruction(BKControl(sub_gate, qubits), qubits)]

        try:
            matrix = protocols.unitary(operation)
            name = str(operation.gate)
            return [
                BKInstruction(
                    braket_gates.Unitary(matrix, display_name=name),
                    qubits,
                )
            ]
        except (ValueError, TypeError) as err:
            raise ProgramConversionError(f"Unable to convert {operation} to Braket") from err

    # Unsupported gates.
    raise ProgramConversionError(f"Unable to convert {operation} to Braket")


def _to_one_qubit_braket_instruction(
    operation: Union[np.ndarray, cirq_ops.Gate, cirq_ops.Operation],
    target: int,
    gate_name: Optional[str] = None,
) -> list[braket.circuits.Instruction]:
    """Converts one-qubit Cirq operation or NumPy array to equivalent Braket instruction(s)

    Args:
        operation: One-qubit Cirq operation or numpy unitary to translate.
        target: Qubit index for the operation to act on. Must be specified and if only
            if `operation` is given as a numpy array.
        gate_name: Optional unitary gate display name for `operation` of type `np.ndarray`

    Raises:
        ValueError: If the operation cannot be converted to Braket.
    """

    def instruction_from_matrix(matrix, target, name):
        display_name = "U" if name is None or "QasmUGate" in name else name
        return [BKInstruction(braket_gates.Unitary(matrix, display_name=display_name), target)]

    def convert_one_qubit_gate(gate, target):
        if isinstance(gate, cirq_ops.XPowGate):
            exponent = gate.exponent
            if np.isclose(exponent, 1.0) or np.isclose(exponent, -1.0):
                return [BKInstruction(braket_gates.X(), target)]
            if np.isclose(exponent, 0.5):
                return [BKInstruction(braket_gates.V(), target)]
            if np.isclose(exponent, -0.5):
                return [BKInstruction(braket_gates.Vi(), target)]

            return [BKInstruction(braket_gates.Rx(exponent * np.pi), target)]

        if isinstance(gate, cirq_ops.YPowGate):
            exponent = gate.exponent

            if np.isclose(exponent, 1.0) or np.isclose(exponent, -1.0):
                return [BKInstruction(braket_gates.Y(), target)]

            return [BKInstruction(braket_gates.Ry(exponent * np.pi), target)]

        if isinstance(gate, cirq_ops.ZPowGate):
            global_shift = gate.global_shift
            exponent = gate.exponent

            if np.isclose(global_shift, 0.0):
                if np.isclose(exponent, 1.0) or np.isclose(exponent, -1.0):
                    return [BKInstruction(braket_gates.Z(), target)]
                if np.isclose(exponent, 0.5):
                    return [BKInstruction(braket_gates.S(), target)]
                if np.isclose(exponent, -0.5):
                    return [BKInstruction(braket_gates.Si(), target)]
                if np.isclose(exponent, 0.25):
                    return [BKInstruction(braket_gates.T(), target)]
                if np.isclose(exponent, -0.25):
                    return [BKInstruction(braket_gates.Ti(), target)]
                return [BKInstruction(braket_gates.PhaseShift(exponent * np.pi), target)]
            if np.isclose(global_shift, -0.5):
                return [BKInstruction(braket_gates.Rz(exponent * np.pi), target)]

        if isinstance(gate, cirq_ops.HPowGate) and np.isclose(abs(gate.exponent), 1.0):
            return [BKInstruction(braket_gates.H(), target)]

        if isinstance(gate, cirq_ops.IdentityGate):
            return [BKInstruction(braket_gates.I(), target)]

        if isinstance(gate, cirq_ops.BitFlipChannel):
            return [BKInstruction(braket_noise_gate.BitFlip(operation.gate._p), target)]

        if isinstance(gate, cirq_ops.PhaseFlipChannel):
            return [BKInstruction(braket_noise_gate.PhaseFlip(operation.gate._p), target)]

        if isinstance(gate, cirq_ops.DepolarizingChannel):
            return [BKInstruction(braket_noise_gate.Depolarizing(operation.gate._p), target)]

        if isinstance(gate, cirq_ops.AmplitudeDampingChannel):
            return [
                BKInstruction(braket_noise_gate.AmplitudeDamping(operation.gate._gamma), target)
            ]

        if isinstance(gate, cirq_ops.GeneralizedAmplitudeDampingChannel):
            return [
                BKInstruction(
                    braket_noise_gate.GeneralizedAmplitudeDamping(
                        gamma=operation.gate._gamma, probability=operation.gate._p
                    ),
                    target,
                )
            ]

        if isinstance(gate, cirq_ops.PhaseDampingChannel):
            return [BKInstruction(braket_noise_gate.PhaseDamping(operation.gate._gamma), target)]

        if cirq_ionq_ops and isinstance(
            gate, (cirq_ionq_ops.GPIGate, cirq_ionq_ops.GPI2Gate, cirq_ionq_ops.MSGate)
        ):
            if isinstance(gate, cirq_ionq_ops.GPIGate):
                return [
                    BKInstruction(braket_gates.GPi(angle=operation.gate.phi * 2 * np.pi), target)
                ]
            if isinstance(gate, cirq_ionq_ops.GPI2Gate):
                return [
                    BKInstruction(braket_gates.GPi2(angle=operation.gate.phi * 2 * np.pi), target)
                ]

        matrix = protocols.unitary(gate)
        gate_name = "U" if isinstance(gate, cirq_ops.MatrixGate) else str(gate)
        return _to_one_qubit_braket_instruction(matrix, target, gate_name=gate_name)

    if isinstance(operation, np.ndarray):
        return instruction_from_matrix(operation, target, gate_name)

    if isinstance(operation, cirq_ops.Operation):
        gate = operation.gate
    elif isinstance(operation, cirq_ops.Gate):
        gate = operation
    else:
        raise ValueError(f"Unable to convert {operation} to braket")

    return convert_one_qubit_gate(gate, target)


def _to_two_qubit_braket_instruction(
    operation: Union[cirq_ops.Gate, cirq_ops.Operation],
    qubits: list[int],
) -> list[braket.circuits.Instruction]:
    """Converts two-qubit Cirq operation to equivalent Braket instruction(s)

    Args:
        operation: Two-qubit Cirq operation to translate.
        qubits: Length 2 list of qubit indicies

    Raises:
        ValueError: If the operation cannot be converted to Braket.
    """
    if isinstance(operation, cirq_ops.Operation):
        gate = operation.gate

    elif isinstance(operation, cirq_ops.Gate):
        gate = operation

    else:
        raise ValueError(f"Unable to convert {operation} to Braket")

    # Translate qubit indices.
    q1, q2 = qubits

    # Check common two-qubit gates.
    if isinstance(gate, cirq_ops.CNotPowGate) and np.isclose(abs(gate.exponent), 1.0):
        return [BKInstruction(braket_gates.CNot(), [q1, q2])]
    if isinstance(gate, cirq_ops.CZPowGate) and np.isclose(abs(gate.exponent), 1.0):
        return [BKInstruction(braket_gates.CZ(), [q1, q2])]
    if isinstance(gate, cirq_ops.SwapPowGate) and np.isclose(gate.exponent, 1.0):
        return [BKInstruction(braket_gates.Swap(), [q1, q2])]
    if isinstance(gate, cirq_ops.ISwapPowGate) and np.isclose(gate.exponent, 1.0):
        return [BKInstruction(braket_gates.ISwap(), [q1, q2])]
    if isinstance(gate, cirq_ops.XXPowGate):
        return [BKInstruction(braket_gates.XX(gate.exponent * np.pi), [q1, q2])]
    if isinstance(gate, cirq_ops.YYPowGate):
        return [BKInstruction(braket_gates.YY(gate.exponent * np.pi), [q1, q2])]
    if isinstance(gate, cirq_ops.ZZPowGate):
        return [BKInstruction(braket_gates.ZZ(gate.exponent * np.pi), [q1, q2])]
    if isinstance(gate, cirq_ops.ControlledGate):
        sub_gate_instr = _to_one_qubit_braket_instruction(gate.sub_gate, q2)
        sub_gate = sub_gate_instr[0].operator
        # return [BKInstruction(sub_gate, target=q2, control=q1)]
        return [BKInstruction(BKControl(sub_gate, [0, 1]), [q1, q2])]
    if isinstance(gate, cirq_ops.DepolarizingChannel):
        return BKInstruction(braket_noise_gate.TwoQubitDepolarizing(operation.gate.p), [q1, q2])
    if isinstance(gate, cirq_ops.KrausChannel):
        return BKInstruction(braket_noise_gate.Kraus(matrices=operation.gate._kraus_ops), [q1, q2])
    if isinstance(gate, cirq_ionq_ops.MSGate):
        return [
            BKInstruction(
                braket_gates.MS(
                    angle_1=operation.gate.phi0 * 2 * np.pi,
                    angle_2=operation.gate.phi1 * 2 * np.pi,
                    angle_3=operation.gate.theta * 2 * np.pi,
                ),
                [q1, q2],
            )
        ]

    # Fallback: arbitrary two-qubit unitary (KAK) decomposition
    unitary = protocols.unitary(operation)
    return _kak_decomposition_to_braket_instruction(unitary, q1, q2)


def _kak_decomposition_to_braket_instruction(
    matrix: np.ndarray, q1: int, q2: int
) -> list[braket.circuits.Instruction]:
    """Converts 4x4 Numpy array to equivalent Braket instruction(s) via kak decomposition

    Args:
        matrix: Unitary 4x4 numpy array representing 2-qubit gate.
        q1: Index of first qubit to act on
        q2: Index of second qubit to act on
    """
    kak = kak_decomposition(matrix)
    A1, A2 = kak.single_qubit_operations_before

    x, y, z = kak.interaction_coefficients
    a = x * -2 / np.pi + 0.5
    b = y * -2 / np.pi + 0.5
    c = z * -2 / np.pi + 0.5

    B1, B2 = kak.single_qubit_operations_after

    return [
        *_to_one_qubit_braket_instruction(A1, q1),
        *_to_one_qubit_braket_instruction(A2, q2),
        BKInstruction(braket_gates.Rx(0.5 * np.pi), q1),
        BKInstruction(braket_gates.CNot(), [q1, q2]),
        BKInstruction(braket_gates.Rx(a * np.pi), q1),
        BKInstruction(braket_gates.Ry(b * np.pi), q2),
        BKInstruction(braket_gates.CNot(), [q2, q1]),
        BKInstruction(braket_gates.Rx(-0.5 * np.pi), q2),
        BKInstruction(braket_gates.Rz(c * np.pi), q2),
        BKInstruction(braket_gates.CNot(), [q1, q2]),
        *_to_one_qubit_braket_instruction(B1, q1),
        *_to_one_qubit_braket_instruction(B2, q2),
    ]



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining Azure Quantum device class for all devices managed by Azure Quantum.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from qbraid.runtime.device import QuantumDevice
from qbraid.runtime.enums import DeviceStatus

from .job import AzureQuantumJob

if TYPE_CHECKING:
    import azure.quantum

    import qbraid.programs
    import qbraid.runtime


class AzureQuantumDevice(QuantumDevice):
    """Azure quantum device interface."""

    def __init__(
        self,
        profile: qbraid.runtime.TargetProfile,
        workspace: azure.quantum.Workspace,
    ):
        super().__init__(profile=profile)
        self._workspace = workspace
        self._device = self.workspace.get_targets(name=self.id)

    @property
    def workspace(self) -> azure.quantum.Workspace:
        """Return the Azure Quantum Workspace."""
        return self._workspace

    def __str__(self):
        """String representation of the AzureQuantumDevice object."""
        return f"{self.__class__.__name__}('{self._device.name}')"

    def status(self) -> qbraid.runtime.enums.DeviceStatus:
        """Return the current status of the Azure device.

        Returns:
            DeviceStatus: The current status of the device.
        """
        status = self._device._current_availability
        status_map = {
            "Available": DeviceStatus.ONLINE,
            "Deprecated": DeviceStatus.UNAVAILABLE,
            "Unavailable": DeviceStatus.OFFLINE,
            "Degraded": DeviceStatus.OFFLINE,
        }
        return status_map.get(status, DeviceStatus.UNAVAILABLE)

    def is_available(self) -> bool:
        """Check if the Azure device is available for use.

        Returns:
            bool: True if the device is available, False otherwise.
        """
        current_status = self.status()
        if current_status == DeviceStatus.ONLINE:
            return True
        return False

    def submit(
        self, run_input: qbraid.programs.QPROGRAM | list(qbraid.programs.QPROGRAM), *args, **kwargs
    ) -> AzureQuantumJob | list(AzureQuantumJob):
        """Submit a job to the Azure device.

        Args:
            run_input (Any): The program to submit.

        Returns:
            AzureQuantumJob: The submitted job.
        """
        if isinstance(run_input, list):
            quantum_jobs = []
            for _, qprogram in enumerate(run_input):
                job = self._device.submit(qprogram, *args, **kwargs)
                quantum_jobs.append(
                    AzureQuantumJob(job_id=job.id, workspace=self.workspace, device=self)
                )
            return quantum_jobs

        job = self._device.submit(run_input, *args, **kwargs)
        return AzureQuantumJob(job_id=job.id, workspace=self.workspace, device=self)
